from collections.abc import Sequence, Mapping

import matplotlib.pyplot as plt
import numpy as np

from scipy.stats import median_abs_deviation

from bayesflow.utils import prepare_plot_data, prettify_subplots, make_quadratic, add_titles_and_labels, add_metric


def recovery(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    point_agg=np.median,
    uncertainty_agg=median_abs_deviation,
    add_corr: bool = True,
    figsize: Sequence[int] = None,
    label_fontsize: int = 16,
    title_fontsize: int = 18,
    metric_fontsize: int = 16,
    tick_fontsize: int = 12,
    color: str = "#132a70",
    num_col: int = None,
    num_row: int = None,
    xlabel: str = "Ground truth",
    ylabel: str = "Estimate",
    **kwargs,
) -> plt.Figure:
    """
    Creates and plots publication-ready recovery plot with true estimate
    vs. point estimate + uncertainty.
    The point estimate can be controlled with the ``point_agg`` argument,
    and the uncertainty estimate can be controlled with the
    ``uncertainty_agg`` argument.

    This plot yields similar information as the "posterior z-score",
    but allows for generic point and uncertainty estimates:

    https://betanalpha.github.io/assets/case_studies/principled_bayesian_workflow.html

    Important:
    Posterior aggregates play no special role in Bayesian inference and should only be used heuristically.
    For instance, in the case of multi-modal posteriors, common point estimates, such as mean, (geometric) median,
    or maximum a posteriori (MAP) mean nothing.

    Parameters
    ----------
    estimates           : np.ndarray of shape (num_datasets, num_post_draws, num_params)
        The posterior draws obtained from num_datasets
    targets        : np.ndarray of shape (num_datasets, num_params)
        The prior draws (true parameters) used for generating the num_datasets
    variable_keys       : list or None, optional, default: None
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names    : list or None, optional, default: None
        The individual parameter names for nice plot titles. Inferred if None
    point_agg         : function to compute point estimates. Default: median
    uncertainty_agg   : function to compute uncertainty estimates. Default: MAD
    add_corr          : boolean, default: True
        Should correlations between estimates and ground truth values be shown?
    figsize           : tuple or None, optional, default : None
        The figure size passed to the matplotlib constructor. Inferred if None.
    label_fontsize    : int, optional, default: 16
        The font size of the y-label text.
    title_fontsize    : int, optional, default: 18
        The font size of the title text.
    metric_fontsize   : int, optional, default: 16
        The font size of the metrics shown as text.
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels.
    color             : str, optional, default: '#8f2727'
        The color for the true vs. estimated scatter points and error bars.
    num_row           : int, optional, default: None
        The number of rows for the subplots. Dynamically determined if None.
    num_col           : int, optional, default: None
        The number of columns for the subplots. Dynamically determined if None.
    xlabel:
    ylabel:

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving

    Raises
    ------
    ShapeError
        If there is a deviation from the expected shapes of ``estimates`` and ``targets``.
    """

    # Gather plot data and metadata into a dictionary
    plot_data = prepare_plot_data(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
        num_col=num_col,
        num_row=num_row,
        figsize=figsize,
    )

    estimates = plot_data.pop("estimates")
    targets = plot_data.pop("targets")

    # Compute point estimates and uncertainties
    point_estimate = point_agg(estimates, axis=1)

    if uncertainty_agg is not None:
        u = uncertainty_agg(estimates, axis=1)

    for i, ax in enumerate(plot_data["axes"].flat):
        if i >= plot_data["num_variables"]:
            break

        # Add scatter and error bars
        if uncertainty_agg is not None:
            _ = ax.errorbar(
                targets[:, i],
                point_estimate[:, i],
                yerr=u[:, i],
                fmt="o",
                alpha=0.5,
                color=color,
                **kwargs,
            )
        else:
            _ = ax.scatter(targets[:, i], point_estimate[:, i], alpha=0.5, color=color, **kwargs)

        make_quadratic(ax, targets[:, i], point_estimate[:, i])

        if add_corr:
            corr = np.corrcoef(targets[:, i], point_estimate[:, i])[0, 1]
            add_metric(ax=ax, metric_text="$r$", metric_value=corr, metric_fontsize=metric_fontsize)

        ax.set_title(plot_data["variable_names"][i], fontsize=title_fontsize)

    # Add custom schmuck
    prettify_subplots(plot_data["axes"], num_subplots=plot_data["num_variables"], tick_fontsize=tick_fontsize)
    add_titles_and_labels(
        axes=plot_data["axes"],
        num_row=plot_data["num_row"],
        num_col=plot_data["num_col"],
        xlabel=xlabel,
        ylabel=ylabel,
        label_fontsize=label_fontsize,
    )

    plot_data["fig"].tight_layout()
    return plot_data["fig"]



from collections.abc import Sequence

from bayesflow.utils.serialization import serializable, serialize

from .transform import Transform


@serializable
class Drop(Transform):
    """
    Transform to drop variables from further calculation.

    Parameters
    ----------
    keys : sequence of str
        Names of data variables that should be dropped

    Examples
    --------

    >>> import bayesflow as bf
    >>> a = [1, 2, 3, 4]
    >>> b = [[1, 2], [3, 4]]
    >>> c = [[5, 6, 7, 8]]
    >>> dat = dict(a=a, b=b, c=c)
    >>> dat
        {'a': [1, 2, 3, 4], 'b': [[1, 2], [3, 4]], 'c': [[5, 6, 7, 8]]}
    >>> drop = bf.adapters.transforms.Drop(("b", "c"))
    >>> drop.forward(dat)
        {'a': [1, 2, 3, 4]}
    """

    def __init__(self, keys: Sequence[str]):
        self.keys = keys

    def get_config(self) -> dict:
        return serialize({"keys": self.keys})

    def forward(self, data: dict[str, any], **kwargs) -> dict[str, any]:
        # no strict version because there is no requirement for the keys to be present
        return {key: value for key, value in data.items() if key not in self.keys}

    def inverse(self, data: dict[str, any], **kwargs) -> dict[str, any]:
        # non-invertible transform
        return data

    def extra_repr(self) -> str:
        return "[" + ", ".join(map(repr, self.keys)) + "]"



import keras


def assert_models_equal(model1: keras.Model, model2: keras.Model):
    assert isinstance(model1, keras.Model)
    assert isinstance(model2, keras.Model)

    for layer1, layer2 in zip(model1.layers, model2.layers):
        if isinstance(layer1, keras.Model):
            assert_models_equal(layer1, layer2)
        else:
            assert_layers_equal(layer1, layer2)


def assert_layers_equal(layer1: keras.Layer, layer2: keras.Layer):
    msg = f"Layers {layer1.name} and {layer2.name} have different types."
    assert type(layer1) is type(layer2), msg

    msg = (
        f"Layers {layer1.name} and {layer2.name} have a different number of variables "
        f"({len(layer1.variables)}, {len(layer2.variables)})."
    )
    assert len(layer1.variables) == len(layer2.variables), msg

    for v1, v2 in zip(layer1.variables, layer2.variables):
        if v1.name == "seed_generator_state":
            # keras issue: https://github.com/keras-team/keras/issues/19796
            continue

        x1 = keras.ops.convert_to_numpy(v1)
        x2 = keras.ops.convert_to_numpy(v2)
        msg = f"Variable '{v1.name}' for Layer '{layer1.name}' is not equal: {x1} != {x2}"
        assert keras.ops.all(keras.ops.isclose(x1, x2)), msg

    # this is turned off for now, see https://github.com/bayesflow-org/bayesflow/issues/412
    msg = f"Layers {layer1.name} and {layer2.name} have a different name."
    # assert layer1.name == layer2.name, msg



from functools import singledispatch
import keras


@singledispatch
def find_network(arg, *args, **kwargs):
    raise TypeError(f"Cannot infer network from {arg!r}.")


@find_network.register
def _(name: str, *args, **kwargs):
    match name.lower():
        case "mlp" | "default":
            from bayesflow.networks import MLP

            network = MLP(*args, **kwargs)
        case other:
            raise ValueError(f"Unsupported network name: '{other}'.")

    return network


@find_network.register
def _(cls: type, *args, **kwargs):
    # Instantiate class with the given arguments
    network = cls(*args, **kwargs)
    return network


@find_network.register
def _(layer: keras.Layer, *args, **kwargs):
    return layer



from keras.saving import register_keras_serializable as serializable

from .normed_difference_score import NormedDifferenceScore


@serializable(package="bayesflow.scores")
class MedianScore(NormedDifferenceScore):
    r""":math:`S(\hat \theta, \theta) = | \hat \theta - \theta |`

    Scores a predicted median with the absolute error score.
    """

    def __init__(self, **kwargs):
        super().__init__(k=1, **kwargs)
        self.config = {}



import keras

from bayesflow.types import Tensor
from bayesflow.utils import layer_kwargs
from bayesflow.utils.decorators import sanitize_input_shape
from bayesflow.utils.serialization import serializable


@serializable
class MambaBlock(keras.Layer):
    """
    Wraps the original Mamba module from, with added functionality for bidirectional processing:
    https://github.com/state-spaces/mamba/blob/main/mamba_ssm/modules/mamba_simple.py

    Copyright (c) 2023, Tri Dao, Albert Gu.
    """

    def __init__(
        self,
        state_dim: int,
        conv_dim: int,
        feature_dim: int = 16,
        expand: int = 2,
        bidirectional: bool = True,
        dt_min: float = 0.001,
        dt_max: float = 0.1,
        device: str = "cuda",
        **kwargs,
    ):
        """
        A Keras layer implementing a Mamba-based sequence processing block.

        This layer applies a Mamba model for sequence modeling, preceded by a
        convolutional projection and followed by layer normalization.

        Parameters
        ----------
        state_dim : int
            The dimension of the state space in the Mamba model.
        conv_dim : int
            The dimension of the convolutional layer used in Mamba.
        feature_dim : int, optional
            The feature dimension for input projection and Mamba processing (default is 16).
        expand : int, optional
            Expansion factor for Mamba's internal dimension (default is 1).
        dt_min : float, optional
            Minimum delta time for Mamba (default is 0.001).
        dt_max : float, optional
            Maximum delta time for Mamba (default is 0.1).
        device : str, optional
            The device to which the Mamba model is moved, typically "cuda" or "cpu" (default is "cuda").
        **kwargs :
            Additional keyword arguments passed to the `keras.layers.Layer` initializer.
        """

        super().__init__(**layer_kwargs(kwargs))

        if keras.backend.backend() != "torch":
            raise RuntimeError("Mamba is only available using torch backend.")

        try:
            from mamba_ssm import Mamba
        except ImportError as e:
            raise ImportError("Could not import Mamba. Please install it via `pip install mamba-ssm`") from e

        self.bidirectional = bidirectional

        self.mamba = Mamba(
            d_model=feature_dim, d_state=state_dim, d_conv=conv_dim, expand=expand, dt_min=dt_min, dt_max=dt_max
        ).to(device)

        self.input_projector = keras.layers.Conv1D(
            feature_dim,
            kernel_size=1,
            strides=1,
        )
        self.layer_norm = keras.layers.LayerNormalization()

    def call(self, x: Tensor, training: bool = False, **kwargs) -> Tensor:
        """
        Applies the Mamba layer to the input tensor `x`, optionally in a bidirectional manner.

        Parameters
        ----------
        x : Tensor
            Input tensor of shape `(batch_size, sequence_length, input_dim)`.
        training : bool, optional
            Whether the layer should behave in training mode (e.g., applying dropout). Default is False.
        **kwargs : dict
            Additional keyword arguments passed to the internal `_call` method.

        Returns
        -------
        Tensor
            Output tensor of shape `(batch_size, sequence_length, feature_dim)` if unidirectional,
            or `(batch_size, sequence_length, 2 * feature_dim)` if bidirectional.
        """

        out_forward = self._call(x, training=training, **kwargs)
        if self.bidirectional:
            out_backward = self._call(keras.ops.flip(x, axis=-2), training=training, **kwargs)
            return keras.ops.concatenate((out_forward, out_backward), axis=-1)
        return out_forward

    def _call(self, x: Tensor, training: bool = False, **kwargs) -> Tensor:
        x = self.input_projector(x)
        h = self.mamba(x)
        out = self.layer_norm(h + x, training=training, **kwargs)
        return out

    @sanitize_input_shape
    def build(self, input_shape):
        super().build(input_shape)
        self.call(keras.ops.zeros(input_shape))



import keras
from keras import layers

from bayesflow.networks import MLP
from bayesflow.types import Tensor
from bayesflow.utils import layer_kwargs
from bayesflow.utils.decorators import sanitize_input_shape
from bayesflow.utils.serialization import serializable


@serializable
class MultiHeadAttentionBlock(keras.Layer):
    """Implements the MAB block from [1] which represents learnable cross-attention.

    In particular, it uses a so-called "Post-LN" transformer block [2] which applies
    layer norm following attention and following MLP. A "Pre-LN" transformer block
    can easily be implemented.

    [1] Lee, J., Lee, Y., Kim, J., Kosiorek, A., Choi, S., & Teh, Y. W. (2019).
        Set transformer: A framework for attention-based permutation-invariant neural networks.
        In International conference on machine learning (pp. 3744-3753). PMLR.

    [2] Xiong, R., Yang, Y., He, D., Zheng, K., Zheng, S., Xing, C., ... & Liu, T. (2020, November).
    On layer normalization in the transformer architecture.
    In International conference on machine learning (pp. 10524-10533). PMLR.
    """

    def __init__(
        self,
        embed_dim: int = 64,
        num_heads: int = 4,
        dropout: float = 0.05,
        mlp_depth: int = 2,
        mlp_width: int = 128,
        mlp_activation: str = "gelu",
        kernel_initializer: str = "lecun_normal",
        use_bias: bool = True,
        layer_norm: bool = True,
        **kwargs,
    ):
        """Creates a multi-head attention block which will typically be used as part of a
        set transformer architecture according to [1]. Corresponds to standard cross-attention.

        Parameters
        ----------
        embed_dim : int, optional
            Dimensionality of the embedding space, by default 64.
        num_heads : int, optional
            Number of attention heads, by default 4.
        dropout : float, optional
            Dropout rate applied to attention and MLP layers, by default 0.05.
        mlp_depth : int, optional
            Number of layers in the feedforward MLP block, by default 2.
        mlp_width : int, optional
            Width of each hidden layer in the MLP block, by default 128.
        mlp_activation : str, optional
            Activation function used in the MLP block, by default "gelu".
        kernel_initializer : str, optional
            Initializer for kernel weights, by default "lecun_normal".
        use_bias : bool, optional
            Whether to include bias terms in dense layers, by default True.
        layer_norm : bool, optional
            Whether to apply layer normalization before and after attention, by default True.
        **kwargs : dict
            Additional keyword arguments passed to the Keras Layer base class.
        """

        super().__init__(**layer_kwargs(kwargs))

        self.input_projector = layers.Dense(embed_dim, name="input_projector")
        self.attention = layers.MultiHeadAttention(
            key_dim=embed_dim,
            num_heads=num_heads,
            dropout=dropout,
            use_bias=use_bias,
            output_shape=embed_dim,
            name="attention",
        )
        self.ln_pre = layers.LayerNormalization(name="layer_norm_pre") if layer_norm else None
        self.mlp = MLP(
            widths=(mlp_width,) * mlp_depth,
            activation=mlp_activation,
            kernel_initializer=kernel_initializer,
            dropout=dropout,
        )
        self.output_projector = layers.Dense(embed_dim, name="output_projector")
        self.ln_post = layers.LayerNormalization(name="layer_norm_post") if layer_norm else None

    def call(self, seq_x: Tensor, seq_y: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Performs the forward pass through the attention layer.

        Parameters
        ----------
        seq_x    : Tensor (e.g., np.ndarray, tf.Tensor, ...)
            Input of shape (batch_size, seq_size_x, input_dim), which will
            play the role of a query (Q).
        seq_y    : Tensor
            Input of shape (batch_size, seq_size_y, input_dim), which will
            play the role of key (K) and value (V).
        training : boolean, optional (default - True)
            Passed to the optional internal dropout and spectral normalization
            layers to distinguish between train and test time behavior.
        **kwargs : dict, optional (default - {})
            Additional keyword arguments passed to the internal attention layer,
            such as ``attention_mask`` or ``return_attention_scores``

        Returns
        -------
        out : Tensor
            Output of shape (batch_size, set_size_x, output_dim)
        """

        h = self.input_projector(seq_x) + self.attention(
            query=seq_x, key=seq_y, value=seq_y, training=training, **kwargs
        )
        if self.ln_pre is not None:
            h = self.ln_pre(h, training=training)

        out = h + self.output_projector(self.mlp(h, training=training))
        if self.ln_post is not None:
            out = self.ln_post(out, training=training)

        return out

    # noinspection PyMethodOverriding
    @sanitize_input_shape
    def build(self, seq_x_shape, seq_y_shape):
        self.call(keras.ops.zeros(seq_x_shape), keras.ops.zeros(seq_y_shape))

    @sanitize_input_shape
    def compute_output_shape(self, seq_x_shape, seq_y_shape):
        return keras.ops.shape(self.call(keras.ops.zeros(seq_x_shape), keras.ops.zeros(seq_y_shape)))



def test_scheduled_integration():
    import keras
    from bayesflow.utils import integrate

    def fn(t, x):
        return {"x": t**2}

    steps = keras.ops.convert_to_tensor([0.0, 0.5, 1.0])
    approximate_result = 0.0 + 0.5**2 * 0.5
    result = integrate(fn, {"x": 0.0}, steps=steps)["x"]
    assert result == approximate_result



from collections.abc import Sequence

import numpy as np

from bayesflow.utils.serialization import serialize, serializable

from .transform import Transform


@serializable
class Concatenate(Transform):
    """Concatenate multiple arrays into a new key. Used to specify how data variables should be treated by the network.

    Parameters
    ----------
    keys : sequence of str,
        Input a list of strings, where the strings are the names of data variables.
    into : str
        A string telling the network how to use the variables named in keys.
    axis : int, optional
        Along which axis to concatenate the keys. The last axis is used by default.

    Examples
    --------
    Suppose you have a simulator that generates variables "beta" and "sigma" from priors and then observation
    variables "x" and "y". We can then use concatenate in the following way

    >>> adapter = (
        bf.Adapter()
            .concatenate(["beta", "sigma"], into="inference_variables")
            .concatenate(["x", "y"], into="summary_variables")
    )
    """

    def __init__(self, keys: Sequence[str], *, into: str, axis: int = -1, indices: list | None = None):
        self.keys = keys
        self.into = into
        self.axis = axis

        self.indices = indices

    def get_config(self) -> dict:
        config = {
            "keys": self.keys,
            "into": self.into,
            "axis": self.axis,
            "indices": self.indices,
        }
        return serialize(config)

    def forward(self, data: dict[str, any], *, strict: bool = True, **kwargs) -> dict[str, any]:
        if not strict and self.indices is None:
            raise ValueError("Cannot call `forward` with `strict=False` before calling `forward` with `strict=True`.")

        # copy to avoid side effects
        data = data.copy()

        required_keys = set(self.keys)
        available_keys = set(data.keys())
        common_keys = available_keys & required_keys
        missing_keys = required_keys - available_keys

        if strict and missing_keys:
            # invalid call
            raise KeyError(f"Missing keys: {missing_keys!r}")
        elif missing_keys:
            # we cannot produce a result, but should still remove the keys
            for key in common_keys:
                data.pop(key)

            return data

        if self.indices is None:
            # remember the indices of the parts in the concatenated array
            self.indices = np.cumsum([data[key].shape[self.axis] for key in self.keys]).tolist()

        # remove each part
        parts = [data.pop(key) for key in self.keys]

        # concatenate them all
        result = np.concatenate(parts, axis=self.axis)

        # store the result
        data[self.into] = result

        return data

    def inverse(self, data: dict[str, any], *, strict: bool = False, **kwargs) -> dict[str, any]:
        if self.indices is None:
            raise RuntimeError("Cannot call `inverse` before calling `forward` at least once.")

        # copy to avoid side effects
        data = data.copy()

        if strict and self.into not in data:
            # invalid call
            raise KeyError(f"Missing key: {self.into!r}")
        elif self.into not in data:
            # nothing to do
            return data

        # split the concatenated array and remove the concatenated key
        keys = self.keys
        values = np.split(data.pop(self.into), self.indices, axis=self.axis)

        # restore the parts
        data |= dict(zip(keys, values))

        return data

    def extra_repr(self) -> str:
        result = "[" + ", ".join(map(repr, self.keys)) + "] -> " + repr(self.into)

        if self.axis != -1:
            result += f", axis={self.axis}"

        return result



import keras

from tests.utils import assert_allclose


def fn(x):
    return keras.ops.square(x)


def test_vjp():
    from bayesflow.utils import vjp

    inputs = keras.random.normal((16, 32))
    tangents = keras.ops.ones((16, 32))

    vjp_fn = vjp(fn, inputs)
    v = vjp_fn(tangents)

    assert keras.ops.shape(v) == (16, 32)
    assert_allclose(v, 2.0 * inputs)


def test_jvp():
    from bayesflow.utils import jvp

    inputs = keras.random.normal((16, 32))
    tangents = keras.ops.ones((16, 32))

    v = jvp(fn, inputs, tangents)

    assert keras.ops.shape(v) == (16, 32)
    assert_allclose(v, 2.0 * inputs)


def test_jacobian():
    from bayesflow.utils import jacobian

    inputs = keras.random.normal((16, 32))

    j = jacobian(fn, inputs)
    target = 2.0 * keras.ops.tile(keras.ops.expand_dims(inputs, axis=-1), (1, 1, 32)) * keras.ops.eye(32)

    assert keras.ops.shape(j) == (16, 32, 32)
    assert_allclose(j, target, atol=0.01, rtol=0.01)


def test_jacobian_trace():
    from bayesflow.utils import jacobian, jacobian_trace

    inputs = keras.random.normal((16, 128))

    # deterministic
    j = jacobian(fn, inputs)
    jt = jacobian_trace(fn, inputs)

    assert jt.shape == (16,)

    jt_target = keras.ops.trace(j, axis1=-2, axis2=-1)

    assert_allclose(jt, jt_target)

    # too few max_steps, uses the stochastic version
    jt = jacobian_trace(fn, inputs, max_steps=127)

    # this check is not reliable enough yet
    # assert_allclose(jt, jt_target, atol=0.01, rtol=0.01)



"""
A collection of :py:class:`~bayesflow.distributions.Distribution`\ s,
which represent the latent space for :py:class:`~bayesflow.networks.InferenceNetwork`\ s
or the summary space of :py:class:`~bayesflow.networks.SummaryNetwork`\ s.
"""

from .distribution import Distribution
from .diagonal_normal import DiagonalNormal
from .diagonal_student_t import DiagonalStudentT
from .mixture import Mixture

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



import keras.callbacks


class FitInterruptedError(Exception):
    pass


class InterruptFitCallback(keras.callbacks.Callback):
    def __init__(self, batches=None, epochs=None, error_type=FitInterruptedError):
        super().__init__()
        if batches is None and epochs is None:
            raise ValueError("Either batches or epochs must be specified.")

        self.batches = batches
        self.epochs = epochs
        self.error_type = error_type

    def on_train_batch_end(self, batch, logs=None):
        if self.batches is not None and self.batches <= batch:
            raise self.error_type()

    def on_epoch_end(self, epoch, logs=None):
        if self.epochs is not None and self.epochs <= epoch:
            raise self.error_type()



import keras
import numpy as np
import pytest


def test_invalid_shape_for_ordered_quantiles(ordered_quantiles, batch_size, num_quantiles, num_variables):
    with pytest.raises(AssertionError) as excinfo:
        ordered_quantiles.build((batch_size, batch_size, num_quantiles, num_variables))

    assert "resolve which axis should be ordered automatically" in str(excinfo)


@pytest.mark.parametrize("axis", [1, 2])
def test_invalid_shape_for_ordered_quantiles_with_specified_axis(
    ordered_quantiles, axis, batch_size, num_quantiles, num_variables
):
    ordered_quantiles.axis = axis
    ordered_quantiles.build((batch_size, batch_size, num_quantiles, num_variables))


def check_ordering(output, axis):
    output = keras.ops.convert_to_numpy(output)
    assert np.all(np.diff(output, axis=axis) > 0), f"is not ordered along specified axis: {axis}."
    for i in range(output.ndim):
        if i != axis % output.ndim:
            assert not np.all(np.diff(output, axis=i) > 0), (
                f"is ordered along axis which is not meant to be ordered: {i}."
            )


@pytest.mark.parametrize("axis", [0, 1, 2])
def test_ordering(axis, unordered):
    from bayesflow.links import Ordered

    activation = Ordered(axis=axis, anchor_index=5)

    output = activation(unordered)

    check_ordering(output, axis)


def test_quantile_ordering(quantiles, unordered):
    from bayesflow.links import OrderedQuantiles

    activation = OrderedQuantiles(q=quantiles)

    activation.build(unordered.shape)
    axis = activation.axis

    output = activation(unordered)

    check_ordering(output, axis)


def test_positive_definite(positive_definite, batch_size, num_variables):
    input_shape = positive_definite.compute_input_shape((batch_size, num_variables, num_variables))

    # Too strongly negative values lead to numerical instabilities -> reduce scale
    random_preactivation = keras.random.normal(input_shape) * 0.1
    output = positive_definite(random_preactivation)
    output = keras.ops.convert_to_numpy(output)

    # Check if output is invertible
    np.linalg.inv(output)

    # Calculated eigenvalues to test for positive definiteness
    eigenvalues = np.linalg.eig(output).eigenvalues

    assert np.all(eigenvalues.real > 0) and np.all(np.isclose(eigenvalues.imag, 0)), (
        f"output is not positive definite: min(real)={np.min(eigenvalues.real)}, "
        f"max(abs(imag))={np.max(np.abs(eigenvalues.imag))}"
    )



import numpy as np

from bayesflow.utils.serialization import serializable

from .elementwise_transform import ElementwiseTransform


@serializable
class AsTimeSeries(ElementwiseTransform):
    """The `.as_time_series` transform can be used to indicate that variables shall be treated as time series.

    Currently, all this transformation does is to ensure that the variable
    arrays are at least 3D. The 2rd dimension is treated as the
    time series dimension and the 3rd dimension as the data dimension.
    In the future, the transform will have more advanced behavior
    to better ensure the correct treatment of time series data.

    Examples
    --------

    >>> adapter = bf.Adapter().as_time_series(["x", "y"])
    """

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return np.atleast_3d(data)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        if data.shape[2] == 1:
            return np.squeeze(data, axis=2)

        return data

    def get_config(self) -> dict:
        return {}



"""
A collection of `keras.Metric <https://keras.io/api/metrics/base_metric/#metric-class>`__\ s for evaluating the
performance of models.
"""

from . import functional
from .maximum_mean_discrepancy import MaximumMeanDiscrepancy
from .root_mean_squard_error import RootMeanSquaredError

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=["functional"])



from numbers import Number

import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform


@serializable
class ToArray(ElementwiseTransform):
    """
    Checks provided data for any non-arrays and converts them to numpy arrays.

    This ensures all data is in a format suitable for training.

    Examples
    --------
    >>> ta = bf.adapters.transforms.ToArray()
    >>> a = [1, 2, 3, 4]
    >>> ta.forward(a)
        array([1, 2, 3, 4])
    >>> b = [[1, 2], [3, 4]]
    >>> ta.forward(b)
        array([[1, 2],
            [3, 4]])
    """

    def __init__(self, original_type: type = None):
        super().__init__()
        self.original_type = original_type

    def get_config(self) -> dict:
        return serialize({"original_type": self.original_type})

    def forward(self, data: any, **kwargs) -> np.ndarray:
        if self.original_type is None:
            self.original_type = type(data)

        return np.asarray(data)

    def inverse(self, data: np.ndarray, **kwargs) -> any:
        if self.original_type is None:
            raise RuntimeError("Cannot call `inverse` before calling `forward` at least once.")

        if issubclass(self.original_type, Number):
            try:
                return self.original_type(data.item())
            except ValueError:
                pass

        # cannot invert
        return data



import keras


def euclidean(x1, x2):
    # TODO: rename and move this function
    result = x1[:, None] - x2[None, :]
    shape = list(keras.ops.shape(result))
    shape[2:] = [-1]
    result = keras.ops.reshape(result, shape)
    result = keras.ops.norm(result, ord=2, axis=-1)
    return result



import asyncio
import json
import logging
import os
from pathlib import Path
import shutil
from subprocess import PIPE, CalledProcessError

from sphinx_polyversion.builder import BuildError
from sphinx_polyversion.driver import DefaultDriver
from sphinx_polyversion.pyvenv import Pip
from sphinx_polyversion.json import JSONable

from typing import Iterator

import tempfile

logging.basicConfig()
logger = logging.getLogger("poly.py")


# adapted from Pip
class DynamicPip(Pip):
    def __init__(self, path: Path, name: str, venv: str | Path, *args, **kwargs):
        """
        Adapt `Pip` to dynamically use name in venv path.

        See `Pip` for detailed docs.
        """
        logger.info("Setting dynamic venv name: " + str(Path(venv) / name))
        super().__init__(path, name, Path(venv) / name, *args, **kwargs)

    async def __aenter__(self):
        """
        Set the venv up.

        Raises
        ------
        BuildError
            Running `pip install` failed.

        """
        await super(Pip, self).__aenter__()

        logger.info("Running `pip install`...")

        cmd: list[str] = ["pip", "install"]
        cmd += self.args

        process = await asyncio.create_subprocess_exec(
            *cmd,
            cwd=self.path,
            env=self.activate(self.apply_overrides(os.environ.copy())),
            stdout=PIPE,
            stderr=PIPE,
        )
        out, err = await process.communicate()
        out = out.decode(errors="ignore")
        err = err.decode(errors="ignore")

        self.logger.debug("Installation output:\n %s", out)
        if process.returncode != 0:
            # see https://github.com/real-yfprojects/sphinx-polyversion/pull/32
            self.logger.error("Installation error:\n %s", err)
            raise BuildError from CalledProcessError(
                returncode=process.returncode, cmd=" ".join(cmd), output=out, stderr=err
            )
        return self


class PyDataVersionEncoder(json.JSONEncoder):
    """Encoder to turn list of built GitRefs into a version.json consumable by Sphinx."""

    def transform(self, o: JSONable):
        output = []
        processed_names = []
        for ref in o:
            if ref.name in processed_names:
                continue
            output.append(
                {
                    "name": ref.name,
                    "version": ref.name,
                    "url": f"/{ref.name}",
                }
            )
            processed_names.append(ref.name)
        # do not use cast for performance reasons
        return output

    def __call__(self, o: JSONable):
        return self.transform(o)

    def iterencode(self, o: JSONable, _one_shot: bool = False) -> Iterator[str]:
        """Encode an object."""
        # called for every top level object to encode
        return super().iterencode(self.transform(o), _one_shot)


class CustomDriver(DefaultDriver):
    """DefaultDriver which does not copy venvs for local builds.

    See https://github.com/real-yfprojects/sphinx-polyversion/pull/31
    """

    async def build_local(self) -> None:
        if not self.mock:
            raise ValueError("Missing mock data.")

        # process mock data
        self.targets = self.builds = self.mock["revisions"]
        rev = self.mock["current"]

        if rev not in self.targets:
            self.targets.append(rev)

        # create builder
        builder = await self.init_builder(rev)

        async def get_unignored_files(directory: Path) -> list[Path]:
            cmd = ["git", "ls-files", "--cached", "--others", "--exclude-standard"]
            process = await asyncio.create_subprocess_exec(*cmd, cwd=directory, stdout=PIPE)
            out, err = await process.communicate()
            files = out.decode().split("\n")
            return [Path(path.strip()) for path in files]

        # create temporary directory to use for building this version
        with tempfile.TemporaryDirectory() as tmp:
            path = Path(tmp)
            # copy source files
            logger.info(f"Copying source files to temporary directory (except for files ignored by git)... '{tmp}'")
            try:
                files = await get_unignored_files(self.root)
                for filename in files:
                    source = self.root / filename
                    target = path / filename
                    if source.exists():
                        target.parent.mkdir(parents=True, exist_ok=True)
                        if not target.exists():
                            shutil.copy2(source, target, follow_symlinks=False)
            except CalledProcessError:
                logger.warning("Could not list un-ignored files using git. Copying full working directory...")
                shutil.copytree(self.root, path, symlinks=True, dirs_exist_ok=True)
            # setup build environment (e.g. poetry/pip venv)
            async with await self.init_environment(path, rev) as env:
                # construct metadata to pass to the build process
                data = await self.init_data(rev, env)
                # build the docs
                artifact = await builder.build(env, self.output_dir / "local", data=data)

        # call hook for success, on failure an exception will have been raised
        self.build_succeeded(rev, artifact)

        await self.build_root()



from collections.abc import Sequence

import numpy as np

import keras
from keras import ops

from bayesflow.types import Shape, Tensor
from bayesflow.utils.decorators import allow_batch_size
from bayesflow.utils.serialization import serializable, serialize
from bayesflow.distributions import Distribution


@serializable
class Mixture(Distribution):
    """Utility class for a backend-agnostic mixture distributions."""

    def __init__(
        self,
        distributions: Sequence[Distribution],
        mixture_logits: Sequence[float] = None,
        trainable_mixture: bool = False,
        **kwargs,
    ):
        """
        Initializes a mixture of distributions as a latent distro.

        Parameters
        ----------
        distributions : Sequence[Distribution]
            A sequence of `Distribution` instances to form the mixture components.
        mixture_logits : Sequence[float], optional
            Initial unnormalized log‑weights for each component. If `None`, all
            components are assigned equal weight. Default is `None`.
        trainable_mixture : bool, optional
            Whether the mixture weights (`mixture_logits`) should be trainable.
            Default is `False`.
        **kwargs
            Additional keyword arguments passed to the base `Distribution` class.

        Attributes
        ----------
        distributions : Sequence[Distribution]
            The list of component distributions.
        mixture_logits : Tensor
            Trainable or fixed logits representing the mixture weights.
        dim : int or None
            Dimensionality of the output samples; set when first sampling.
        """

        super().__init__(**kwargs)

        self.distributions = distributions

        if mixture_logits is None:
            self.mixture_logits = ops.ones(shape=len(distributions))
        else:
            self.mixture_logits = ops.convert_to_tensor(mixture_logits)

        self.trainable_mixture = trainable_mixture

        self.dim = None
        self._mixture_logits = None

    @allow_batch_size
    def sample(self, batch_shape: Shape) -> Tensor:
        """
        Draws samples from the mixture distribution by sampling a categorical index
        for each entry in `batch_shape` according to the softmax of `mixture_logits`,
        then draws from the corresponding component distribution.

        Parameters
        ----------
        batch_shape : Shape
            The desired sample batch shape (tuple of ints), not including the
            event dimension.

        Returns
        -------
        samples: Tensor
            A tensor of shape `batch_shape + (dim,)` containing samples drawn
            from the mixture.
        """
        # Will use numpy until keras adds support for N-D categorical sampling
        pvals = keras.ops.convert_to_numpy(keras.ops.softmax(self._mixture_logits))
        cat_samples = np.random.multinomial(n=1, pvals=pvals, size=batch_shape)
        cat_samples = cat_samples.argmax(axis=-1)

        # Prepare array to fill and dtype to infer
        samples = np.zeros(batch_shape + (self.dim,))
        dtype = None

        # Fill in array with vectorized sampling per component
        for i in range(len(self.distributions)):
            dist_mask = cat_samples == i
            dist_indices = np.where(dist_mask)
            num_dist_samples = np.sum(dist_mask)
            dist_samples = keras.ops.convert_to_numpy(self.distributions[i].sample(num_dist_samples))

            samples[dist_indices] = dist_samples

            dtype = dtype or keras.ops.dtype(dist_samples)

        # Convert to keras for compatibility
        samples = keras.ops.convert_to_tensor(samples, dtype=dtype)

        return samples

    def log_prob(self, samples: Tensor, *, normalize: bool = True) -> Tensor:
        """
        Compute the log probability of given samples under the mixture.

        For each input sample, computes the weighted log‑sum‑exp of the component
        log‑probabilities plus the mixture log‑weights.

        Parameters
        ----------
        samples : Tensor
            A tensor of samples with shape `batch_shape + (dim,)`.
        normalize : bool, optional
            If `True`, returns normalized log‑probabilities (i.e., includes the
            log normalization constant). Default is `True`.

        Returns
        -------
        Tensor
            A tensor of shape `batch_shape`, containing the log probability of
            each sample under the mixture distribution.
        """

        log_prob = [distribution.log_prob(samples, normalize=normalize) for distribution in self.distributions]
        log_prob = ops.stack(log_prob, axis=-1)
        log_prob = ops.logsumexp(log_prob + ops.log_softmax(self._mixture_logits), axis=-1)
        return log_prob

    def build(self, input_shape: Shape) -> None:
        if self.built:
            return

        self.dim = input_shape[-1]

        for distribution in self.distributions:
            distribution.build(input_shape)

        self._mixture_logits = self.add_weight(
            shape=(len(self.distributions),),
            initializer=keras.initializers.get(self.mixture_logits),
            dtype="float32",
            trainable=self.trainable_mixture,
        )

    def get_config(self):
        base_config = super().get_config()

        config = {
            "distributions": self.distributions,
            "mixture_logits": self.mixture_logits,
            "trainable_mixture": self.trainable_mixture,
        }

        return base_config | serialize(config)



from collections.abc import Mapping, Sequence

import numpy as np
import matplotlib.pyplot as plt

from bayesflow.utils import (
    prepare_plot_data,
    add_titles_and_labels,
    add_metric,
    prettify_subplots,
)

from bayesflow.diagnostics.metrics import expected_calibration_error


def mc_calibration(
    pred_models: Mapping[str, np.ndarray] | np.ndarray,
    true_models: Mapping[str, np.ndarray] | np.ndarray,
    model_names: Sequence[str] = None,
    num_bins: int = 10,
    label_fontsize: int = 16,
    title_fontsize: int = 18,
    metric_fontsize: int = 14,
    tick_fontsize: int = 12,
    epsilon: float = 0.02,
    figsize: Sequence[int] = None,
    color: str = "#132a70",
    num_col: int = None,
    num_row: int = None,
) -> plt.Figure:
    """Plots the calibration curves, the ECEs and the marginal histograms of predicted posterior model probabilities
    for a model comparison problem. The marginal histograms inform about the fraction of predictions in each bin.
    Depends on the ``expected_calibration_error`` function for computing the ECE.

    Parameters
    ----------
    pred_models       : np.ndarray of shape (num_data_sets, num_models)
        The predicted posterior model probabilities (PMPs) per data set.
    true_models       : np.ndarray of shape (num_data_sets, num_models)
        The one-hot-encoded true model indices per data set.
    model_names       : list or None, optional, default: None
        The model names for nice plot titles. Inferred if None.
    num_bins          : int, optional, default: 10
        The number of bins to use for the calibration curves (and marginal histograms).
    label_fontsize    : int, optional, default: 16
        The font size of the y-label and y-label texts
    metric_fontsize   : int, optional, default: 14
        The font size of the metric (e.g., ECE)
    title_fontsize    : int, optional, default: 18
        The font size of the title text. Only relevant if `stacked=False`
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels
    epsilon           : float, optional, default: 0.02
        A small amount to pad the [0, 1]-bounded axes from both side.
    figsize          : tuple or None, optional, default: None
        The figure size passed to the ``matplotlib`` constructor. Inferred if ``None``
    color             : str, optional, default: '#8f2727'
        The color of the calibration curves
    num_row             : int, optional, default: None
        The number of rows for the subplots. Dynamically determined if None.
    num_col             : int, optional, default: None
        The number of columns for the subplots. Dynamically determined if None.

    Returns
    -------
    fig : plt.Figure - the figure instance for optional saving
    """

    # Gather plot data and metadata into a dictionary
    plot_data = prepare_plot_data(
        estimates=pred_models,
        targets=true_models,
        variable_names=model_names,
        num_col=num_col,
        num_row=num_row,
        figsize=figsize,
        default_name="M",
    )

    # compute ece and probs
    ece = expected_calibration_error(
        estimates=pred_models,
        targets=true_models,
        model_names=plot_data["variable_names"],
        num_bins=num_bins,
        return_probs=True,
    )

    for j, ax in enumerate(plot_data["axes"].flat):
        # Plot calibration curve
        ax.plot(ece["probs_pred"][j], ece["probs_true"][j], "o-", color=color)

        # Plot PMP distribution over bins
        uniform_bins = np.linspace(0.0, 1.0, num_bins + 1)
        norm_weights = np.ones_like(plot_data["estimates"]) / len(plot_data["estimates"])
        ax.hist(plot_data["estimates"][:, j], bins=uniform_bins, weights=norm_weights[:, j], color="grey", alpha=0.3)

        # Plot AB line
        ax.plot((0, 1), (0, 1), "--", color="black", alpha=0.9)

        # Tweak plot
        ax.set_xlim([0 - epsilon, 1 + epsilon])
        ax.set_ylim([0 - epsilon, 1 + epsilon])
        ax.set_xticks([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])
        ax.set_yticks([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])

        # Add ECE label
        add_metric(
            ax,
            metric_text=r"$\widehat{{\mathrm{{ECE}}}}$",
            metric_value=ece["values"][j],
            metric_fontsize=metric_fontsize,
        )

    prettify_subplots(axes=plot_data["axes"], num_subplots=plot_data["num_variables"], tick_fontsize=tick_fontsize)

    # Add labels, titles, and set font sizes
    add_titles_and_labels(
        axes=plot_data["axes"],
        num_row=plot_data["num_row"],
        num_col=plot_data["num_col"],
        title=plot_data["variable_names"],
        xlabel="Predicted Probability",
        ylabel="True Probability",
        title_fontsize=title_fontsize,
        label_fontsize=label_fontsize,
    )

    plot_data["fig"].tight_layout()
    return plot_data["fig"]



import sys
import types


def _add_imports_to_all(include_modules: bool | list[str] = False, exclude: list[str] | None = None):
    """Add all global variables to __all__"""
    if not isinstance(include_modules, (bool, list)):
        raise ValueError("include_modules must be a boolean or a list of strings")

    exclude_set = set(exclude or [])
    contains = exclude_set.__contains__
    mod_type = types.ModuleType
    frame = sys._getframe(1)
    g: dict = frame.f_globals
    existing = set(g.get("__all__", []))

    to_add = []
    include_list = include_modules if isinstance(include_modules, list) else ()
    inc_all = include_modules is True

    for name, val in g.items():
        if name.startswith("_") or contains(name):
            continue

        if isinstance(val, mod_type):
            if inc_all or name in include_list:
                to_add.append(name)
        else:
            to_add.append(name)

    g["__all__"] = sorted(existing.union(to_add))



import pytest

import numpy as np
from scipy.stats import norm, multivariate_t

import keras

from bayesflow.distributions import DiagonalNormal, DiagonalStudentT, Mixture
from bayesflow.utils.serialization import serialize, deserialize


def test_sample_output_shape(distribution, shape):
    distribution.build(shape)
    samples = distribution.sample(shape[:1])
    assert keras.ops.shape(samples) == shape


def test_log_prob_output_shape(distribution, random_samples):
    distribution.build(keras.ops.shape(random_samples))
    log_prob = distribution.log_prob(random_samples)
    assert keras.ops.shape(log_prob) == keras.ops.shape(random_samples)[:1]


def test_log_prob_correctness(distribution, random_samples):
    distribution.build(keras.ops.shape(random_samples))
    log_prob = distribution.log_prob(random_samples)
    log_prob = keras.ops.convert_to_numpy(log_prob)
    random_samples = keras.ops.convert_to_numpy(random_samples)

    if isinstance(distribution, DiagonalNormal):
        loc = keras.ops.convert_to_numpy(distribution.mean)
        scale = keras.ops.convert_to_numpy(distribution.std)
        log_prob_scipy = norm(loc=loc, scale=scale).logpdf(random_samples)
        log_prob_scipy = log_prob_scipy.sum(axis=-1)

    elif isinstance(distribution, DiagonalStudentT):
        loc = keras.ops.convert_to_numpy(distribution.loc)
        scale = keras.ops.convert_to_numpy(distribution.scale)
        df = distribution.df
        log_prob_scipy = multivariate_t(loc=loc, shape=np.diag(scale**2), df=df).logpdf(random_samples)

    elif isinstance(distribution, Mixture):
        loc = keras.ops.convert_to_numpy(distribution.distributions[0].mean)
        scale = keras.ops.convert_to_numpy(distribution.distributions[0].std)
        log_prob_norm_scipy = norm(loc=loc, scale=scale).logpdf(random_samples)
        log_prob_norm_scipy = log_prob_norm_scipy.sum(axis=-1)

        loc = keras.ops.convert_to_numpy(distribution.distributions[1].loc)
        scale = keras.ops.convert_to_numpy(distribution.distributions[1].scale)
        df = distribution.distributions[1].df
        log_prob_t_scipy = multivariate_t(loc=loc, shape=np.diag(scale**2), df=df).logpdf(random_samples)
        log_prob_scipy = np.log(0.5 * np.exp(log_prob_norm_scipy) + 0.5 * np.exp(log_prob_t_scipy))

    else:
        raise RuntimeError("distribution must be in '[DiagonalNormal, DiagonalStudentT, Mixture]'")

    assert np.allclose(log_prob, log_prob_scipy)


@pytest.mark.parametrize("automatic", [True, False])
def test_build(automatic, distribution, random_samples):
    assert distribution.built is False

    if automatic:
        distribution(random_samples)
    else:
        distribution.build(keras.ops.shape(random_samples))

    assert distribution.built is True


def test_serialize_deserialize(distribution, random_samples):
    distribution.build(keras.ops.shape(random_samples))

    serialized = serialize(distribution)
    deserialized = deserialize(serialized)
    reserialized = serialize(deserialized)

    assert reserialized == serialized



from .as_set import AsSet
from .as_time_series import AsTimeSeries
from .broadcast import Broadcast
from .concatenate import Concatenate
from .constrain import Constrain
from .convert_dtype import ConvertDType
from .drop import Drop
from .elementwise_transform import ElementwiseTransform
from .expand_dims import ExpandDims
from .filter_transform import FilterTransform
from .keep import Keep
from .log import Log
from .map_transform import MapTransform
from .numpy_transform import NumpyTransform
from .one_hot import OneHot
from .rename import Rename
from .scale import Scale
from .serializable_custom_transform import SerializableCustomTransform
from .shift import Shift
from .split import Split
from .sqrt import Sqrt
from .standardize import Standardize
from .to_array import ToArray
from .to_dict import ToDict
from .transform import Transform

from ...utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=["transforms"])



import numpy as np

from bayesflow.simulators import Simulator
from bayesflow.types import Shape, Tensor
from bayesflow.utils.decorators import allow_batch_size


class NormalSimulator(Simulator):
    """TODO: Docstring"""

    @allow_batch_size
    def sample(self, batch_shape: Shape, num_observations: int = 32) -> dict[str, Tensor]:
        mean = np.random.normal(0.0, 0.1, size=batch_shape + (2,))
        std = np.random.lognormal(0.0, 0.1, size=batch_shape + (2,))
        noise = np.random.standard_normal(batch_shape + (num_observations, 2))

        x = mean[:, None] + std[:, None] * noise
        # flatten observations for use without summary network
        x = x.reshape(x.shape[0], -1)

        mean = mean.astype("float32")
        std = std.astype("float32")
        x = x.astype("float32")
        return dict(mean=mean, std=std, x=x)



import pytest


@pytest.fixture()
def inference_network():
    from bayesflow.networks import CouplingFlow

    return CouplingFlow(depth=2, subnet="mlp", subnet_kwargs=dict(widths=(32, 32)))


@pytest.fixture()
def approximator(adapter, inference_network):
    from bayesflow import ContinuousApproximator

    return ContinuousApproximator(
        adapter=adapter,
        inference_network=inference_network,
    )


@pytest.fixture()
def batch_size():
    return 128


@pytest.fixture()
def adapter():
    from bayesflow import ContinuousApproximator

    return ContinuousApproximator.build_adapter(
        inference_variables=["parameters"],
        inference_conditions=["observables"],
    )


@pytest.fixture()
def random_samples(batch_size, simulator):
    return simulator.sample((batch_size,))


@pytest.fixture()
def simulator():
    from bayesflow.simulators import TwoMoons

    return TwoMoons()


@pytest.fixture()
def train_dataset(batch_size, adapter, simulator):
    from bayesflow import OfflineDataset

    num_batches = 32
    data = simulator.sample((num_batches * batch_size,))
    return OfflineDataset(
        data=data,
        adapter=adapter,
        batch_size=batch_size,
        workers=4,
        max_queue_size=num_batches,
        use_multiprocessing=False,
    )


@pytest.fixture()
def validation_dataset(batch_size, adapter, simulator):
    from bayesflow import OfflineDataset

    num_batches = 8
    data = simulator.sample((num_batches * batch_size,))
    return OfflineDataset(data=data, adapter=adapter, batch_size=batch_size, workers=4, max_queue_size=num_batches)



import keras

from bayesflow.utils import model_kwargs
from bayesflow.utils.serialization import deserialize, serializable, serialize
from bayesflow.types import Tensor

from .single_coupling import SingleCoupling

from ..invertible_layer import InvertibleLayer


@serializable
class DualCoupling(InvertibleLayer):
    def __init__(self, subnet: str | type = "mlp", transform: str = "affine", **kwargs):
        super().__init__(**kwargs)
        self.subnet = subnet
        self.transform = transform

        self.coupling1 = SingleCoupling(subnet, transform, **kwargs)
        self.coupling2 = SingleCoupling(subnet, transform, **kwargs)
        self.pivot = None

    def get_config(self):
        base_config = super().get_config()
        base_config = model_kwargs(base_config)

        config = {
            "subnet": self.subnet,
            "transform": self.transform,
        }

        return base_config | serialize(config)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    # noinspection PyMethodOverriding
    def build(self, xz_shape, conditions_shape=None):
        self.pivot = xz_shape[-1] // 2

        xz = keras.ops.zeros(xz_shape)
        if conditions_shape is None:
            conditions = None
        else:
            conditions = keras.ops.zeros(conditions_shape)

        # build nested layers with forward pass
        self.call(xz, conditions=conditions)

    def call(
        self, xz: Tensor, conditions: Tensor = None, inverse: bool = False, training: bool = False, **kwargs
    ) -> (Tensor, Tensor):
        if inverse:
            return self._inverse(xz, conditions=conditions, training=training, **kwargs)
        return self._forward(xz, conditions=conditions, training=training, **kwargs)

    def _forward(self, x: Tensor, conditions: Tensor = None, training: bool = False, **kwargs) -> (Tensor, Tensor):
        """Transform (x1, x2) -> (g(x1; f(x2; x1)), f(x2; x1))"""
        x1, x2 = x[..., : self.pivot], x[..., self.pivot :]
        (z1, z2), log_det1 = self.coupling1(x1, x2, conditions=conditions, training=training, **kwargs)
        (z2, z1), log_det2 = self.coupling2(z2, z1, conditions=conditions, training=training, **kwargs)

        z = keras.ops.concatenate([z1, z2], axis=-1)
        log_det = log_det1 + log_det2

        return z, log_det

    def _inverse(self, z: Tensor, conditions: Tensor = None, training: bool = False, **kwargs) -> (Tensor, Tensor):
        """Transform (g(x1; f(x2; x1)), f(x2; x1)) -> (x1, x2)"""
        z1, z2 = z[..., : self.pivot], z[..., self.pivot :]
        (z2, z1), log_det2 = self.coupling2(z2, z1, conditions=conditions, inverse=True, training=training, **kwargs)
        (x1, x2), log_det1 = self.coupling1(z1, z2, conditions=conditions, inverse=True, training=training, **kwargs)

        x = keras.ops.concatenate([x1, x2], axis=-1)
        log_det = log_det1 + log_det2

        return x, log_det



import pytest
import keras
import numpy as np

from bayesflow.networks.coupling_flow.permutations import (
    FixedPermutation,
    OrthogonalPermutation,
    RandomPermutation,
    Swap,
)


@pytest.fixture(params=[FixedPermutation, OrthogonalPermutation, RandomPermutation, Swap])
def permutation_class(request):
    return request.param


@pytest.fixture
def input_tensor():
    return keras.random.normal((2, 5))


def test_fixed_permutation_build_and_call():
    # Since FixedPermutation is abstract, create a subclass for testing build.
    class TestPerm(FixedPermutation):
        def build(self, xz_shape, **kwargs):
            length = xz_shape[-1]
            self.forward_indices = keras.ops.arange(length - 1, -1, -1)
            self.inverse_indices = keras.ops.arange(length - 1, -1, -1)

    layer = TestPerm()
    input_shape = (2, 4)
    layer.build(input_shape)

    x = keras.ops.convert_to_tensor(np.arange(8).reshape(input_shape).astype("float32"))
    z, log_det = layer(x, inverse=False)
    x_inv, log_det_inv = layer(z, inverse=True)

    # Check shape preservation
    assert z.shape == x.shape
    assert x_inv.shape == x.shape
    # Forward then inverse recovers input
    np.testing.assert_allclose(keras.ops.convert_to_numpy(x_inv), keras.ops.convert_to_numpy(x), atol=1e-5)
    # log_det values should be zero tensors with the correct shape
    assert tuple(log_det.shape) == input_shape[:-1]
    assert tuple(log_det_inv.shape) == input_shape[:-1]


def test_orthogonal_permutation_build_and_call(input_tensor):
    layer = OrthogonalPermutation()
    input_shape = keras.ops.shape(input_tensor)
    layer.build(input_shape)

    z, log_det = layer(input_tensor)
    x_inv, log_det_inv = layer(z, inverse=True)

    # Check output shapes
    assert z.shape == input_tensor.shape
    assert x_inv.shape == input_tensor.shape

    # Forward + inverse should approximately recover input (allow some numeric tolerance)
    np.testing.assert_allclose(
        keras.ops.convert_to_numpy(x_inv), keras.ops.convert_to_numpy(input_tensor), rtol=1e-5, atol=1e-5
    )

    # log_det should be scalar or batched scalar
    if len(log_det.shape) > 0:
        assert log_det.shape[0] == input_tensor.shape[0]  # batch dim
    else:
        assert log_det.shape == ()

    # log_det_inv should be negative of log_det (det(inv) = 1/det)
    log_det_np = keras.ops.convert_to_numpy(log_det)
    log_det_inv_np = keras.ops.convert_to_numpy(log_det_inv)
    np.testing.assert_allclose(log_det_inv_np, -log_det_np, rtol=1e-5, atol=1e-5)


def test_random_permutation_build_and_call(input_tensor):
    layer = RandomPermutation()
    input_shape = keras.ops.shape(input_tensor)
    layer.build(input_shape)

    # Assert forward_indices and inverse_indices are set and consistent
    fwd = keras.ops.convert_to_numpy(layer.forward_indices)
    inv = keras.ops.convert_to_numpy(layer.inverse_indices)
    # Applying inv on fwd must yield ordered indices
    reordered = fwd[inv]
    np.testing.assert_array_equal(np.arange(len(fwd)), reordered)

    z, log_det = layer(input_tensor)
    x_inv, log_det_inv = layer(z, inverse=True)

    assert z.shape == input_tensor.shape
    assert x_inv.shape == input_tensor.shape
    np.testing.assert_allclose(keras.ops.convert_to_numpy(x_inv), keras.ops.convert_to_numpy(input_tensor), atol=1e-5)
    assert tuple(log_det.shape) == input_shape[:-1]
    assert tuple(log_det_inv.shape) == input_shape[:-1]


def test_swap_build_and_call(input_tensor):
    layer = Swap()
    input_shape = keras.ops.shape(input_tensor)
    layer.build(input_shape)

    fwd = keras.ops.convert_to_numpy(layer.forward_indices)
    inv = keras.ops.convert_to_numpy(layer.inverse_indices)
    reordered = fwd[inv]
    np.testing.assert_array_equal(np.arange(len(fwd)), reordered)

    z, log_det = layer(input_tensor)
    x_inv, log_det_inv = layer(z, inverse=True)

    assert z.shape == input_tensor.shape
    assert x_inv.shape == input_tensor.shape
    np.testing.assert_allclose(keras.ops.convert_to_numpy(x_inv), keras.ops.convert_to_numpy(input_tensor), atol=1e-5)
    assert tuple(log_det.shape) == input_shape[:-1]
    assert tuple(log_det_inv.shape) == input_shape[:-1]



import numpy as np
import pytest
from bayesflow.utils.ecdf import fractional_ranks, distance_ranks


@pytest.fixture
def test_data():
    # Provide sample data for testing
    np.random.seed(42)  # Set seed for reproducibility
    post_samples = np.array([[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]])
    prior_samples = np.array([[0.2, 0.3], [0.6, 0.7]])
    references = np.array([[0.0, 0.0], [0.0, 0.0]])
    return post_samples, prior_samples, references


def test_fractional_ranks(test_data):
    post_samples, prior_samples, _ = test_data
    # Compute the expected result manually
    expected = np.mean(post_samples < prior_samples[:, np.newaxis, :], axis=1)
    result = fractional_ranks(post_samples, prior_samples)
    np.testing.assert_almost_equal(result, expected, decimal=6)


@pytest.mark.parametrize("stacked, expected_shape", [(True, (2, 1)), (False, (2, 2))])
def test_distance_ranks(test_data, stacked, expected_shape):
    post_samples, prior_samples, references = test_data
    result = distance_ranks(
        post_samples=post_samples, prior_samples=prior_samples, stacked=stacked, references=references
    )
    assert result.shape == expected_shape


@pytest.mark.parametrize("stacked, expected_shape", [(True, (2, 1)), (False, (2, 2))])
def test_distance_ranks_function(test_data, stacked, expected_shape):
    post_samples, prior_samples, references = test_data

    def test_distance(x, y):
        return np.linalg.norm(x - y)

    result = distance_ranks(
        post_samples=post_samples,
        prior_samples=prior_samples,
        stacked=stacked,
        references=references,
        distance=test_distance,
    )
    assert result.shape == expected_shape



from copy import copy

import builtins
import inspect
import keras
import numpy as np
import sys

# this import needs to be exactly like this to work with monkey patching
from keras.saving import deserialize_keras_object

from .context_managers import monkey_patch
from .decorators import allow_args


PREFIX = "_bayesflow_"

_type_prefix = "__bayesflow_type__"


def serialize_value_or_type(config, name, obj):
    """Serialize an object that can be either a value or a type
    and add it to a copy of the supplied dictionary.

    Parameters
    ----------
    config  : dict
        Dictionary to add the serialized object to. This function does not
        modify the dictionary in place, but returns a modified copy.
    name    : str
        Name of the obj that should be stored. Required for later deserialization.
    obj     : object or type
        The object to serialize. If `obj` is of type `type`, we use
        `keras.saving.get_registered_name` to obtain the registered type name.
        If it is not a type, we try to serialize it as a Keras object.

    Returns
    -------
    updated_config  : dict
        Updated dictionary with a new key `"_bayesflow_<name>_type"` or
        `"_bayesflow_<name>_val"`. The prefix is used to avoid name collisions,
        the suffix indicates how the stored value has to be deserialized.

    Notes
    -----
    We allow strings or `type` parameters at several places to instantiate objects
    of a given type (e.g., `subnet` in `CouplingFlow`). As `type` objects cannot
    be serialized, we have to distinguish the two cases for serialization and
    deserialization. This function is a helper function to standardize and
    simplify this.
    """
    updated_config = config.copy()
    if isinstance(obj, type):
        updated_config[f"{PREFIX}{name}_type"] = keras.saving.get_registered_name(obj)
    else:
        updated_config[f"{PREFIX}{name}_val"] = keras.saving.serialize_keras_object(obj)
    return updated_config


def deserialize_value_or_type(config, name):
    """Deserialize an object that can be either a value or a type and add
    it to the supplied dictionary.

    Parameters
    ----------
    config  : dict
        Dictionary containing the object to deserialize. If a type was
        serialized, it should contain the key `"_bayesflow_<name>_type"`.
        If an object was serialized, it should contain the key
        `"_bayesflow_<name>_val"`. In a copy of this dictionary,
        the item will be replaced with the key `name`.
    name    : str
        Name of the object to deserialize.

    Returns
    -------
    updated_config  : dict
        Updated dictionary with a new key `name`, with a value that is either
        a type or an object.

    See Also
    --------
    serialize_value_or_type
    """
    updated_config = config.copy()
    if f"{PREFIX}{name}_type" in config:
        updated_config[name] = keras.saving.get_registered_object(config[f"{PREFIX}{name}_type"])
        del updated_config[f"{PREFIX}{name}_type"]
    elif f"{PREFIX}{name}_val" in config:
        updated_config[name] = keras.saving.deserialize_keras_object(config[f"{PREFIX}{name}_val"])
        del updated_config[f"{PREFIX}{name}_val"]
    return updated_config


def deserialize(obj, custom_objects=None, safe_mode=True, **kwargs):
    with monkey_patch(deserialize_keras_object, deserialize) as original_deserialize:
        if isinstance(obj, str) and obj.startswith(_type_prefix):
            # we marked this as a type during serialization
            obj = obj[len(_type_prefix) :]
            tp = keras.saving.get_registered_object(
                # TODO: can we pass module objects without overwriting numpy's dict with builtins?
                obj,
                custom_objects=custom_objects,
                module_objects=np.__dict__ | builtins.__dict__,
            )
            if tp is None:
                raise ValueError(
                    f"Could not deserialize type {obj!r}. Make sure it is registered with "
                    f"`keras.saving.register_keras_serializable` or pass it in `custom_objects`."
                )
            return tp
        if inspect.isclass(obj):
            # add this base case since keras does not cover it
            return obj

        obj = original_deserialize(obj, custom_objects=custom_objects, safe_mode=safe_mode, **kwargs)

        return obj


@allow_args
def serializable(cls, package=None, name=None):
    if package is None:
        frame = sys._getframe(1)
        g = frame.f_globals
        package = g.get("__name__", "bayesflow")

    if name is None:
        name = copy(cls.__name__)

    # register subclasses as keras serializable
    return keras.saving.register_keras_serializable(package=package, name=name)(cls)


def serialize(obj):
    if isinstance(obj, (tuple, list, dict)):
        return keras.tree.map_structure(serialize, obj)
    elif inspect.isclass(obj):
        return _type_prefix + keras.saving.get_registered_name(obj)

    return keras.saving.serialize_keras_object(obj)



import pytest

from bayesflow.networks import MLP


@pytest.fixture()
def flow_matching():
    from bayesflow.networks import FlowMatching

    return FlowMatching(
        subnet=MLP([8, 8]),
        integrate_kwargs={"method": "rk45", "steps": 100},
    )


@pytest.fixture()
def consistency_model():
    from bayesflow.networks import ConsistencyModel

    return ConsistencyModel(total_steps=100, subnet=MLP([8, 8]))


@pytest.fixture()
def affine_coupling_flow():
    from bayesflow.networks import CouplingFlow

    return CouplingFlow(
        depth=2, subnet="mlp", subnet_kwargs=dict(widths=[8, 8]), transform="affine", transform_kwargs=dict(clamp=1.8)
    )


@pytest.fixture()
def spline_coupling_flow():
    from bayesflow.networks import CouplingFlow

    return CouplingFlow(
        depth=2, subnet="mlp", subnet_kwargs=dict(widths=[8, 8]), transform="spline", transform_kwargs=dict(bins=8)
    )


@pytest.fixture()
def free_form_flow():
    from bayesflow.experimental import FreeFormFlow

    return FreeFormFlow(encoder_subnet=MLP([16, 16]), decoder_subnet=MLP([16, 16]))


@pytest.fixture()
def typical_point_inference_network():
    from bayesflow.networks import PointInferenceNetwork
    from bayesflow.scores import MeanScore, MedianScore, QuantileScore, MultivariateNormalScore

    return PointInferenceNetwork(
        scores=dict(
            mean=MeanScore(),
            median=MedianScore(),
            quantiles=QuantileScore([0.1, 0.2, 0.5, 0.65]),
            mvn=MultivariateNormalScore(),  # currently not stable
        )
    )


@pytest.fixture()
def typical_point_inference_network_subnet():
    from bayesflow.networks import PointInferenceNetwork
    from bayesflow.scores import MeanScore, MedianScore, QuantileScore, MultivariateNormalScore

    subnet = MLP([16, 8])

    return PointInferenceNetwork(
        scores=dict(
            mean=MeanScore(subnets=dict(value=subnet)),
            median=MedianScore(subnets=dict(value=subnet)),
            quantiles=QuantileScore(subnets=dict(value=subnet)),
            mvn=MultivariateNormalScore(subnets=dict(mean=subnet, covariance=subnet)),
        ),
        subnet=subnet,
    )


@pytest.fixture(
    params=[
        "typical_point_inference_network",
        "affine_coupling_flow",
        "spline_coupling_flow",
        "flow_matching",
        "free_form_flow",
        "consistency_model",
    ],
    scope="function",
)
def inference_network(request):
    return request.getfixturevalue(request.param)


@pytest.fixture(
    params=[
        "typical_point_inference_network_subnet",
        "coupling_flow_subnet",
        "flow_matching_subnet",
        "free_form_flow_subnet",
    ],
    scope="function",
)
def inference_network_subnet(request):
    return request.getfixturevalue(request.param)


@pytest.fixture(
    params=["affine_coupling_flow", "spline_coupling_flow", "flow_matching", "free_form_flow", "consistency_model"],
    scope="function",
)
def generative_inference_network(request):
    return request.getfixturevalue(request.param)


@pytest.fixture(scope="function")
def time_series_network(summary_dim):
    from bayesflow.networks import TimeSeriesNetwork

    return TimeSeriesNetwork(summary_dim=summary_dim)


@pytest.fixture(scope="function")
def time_series_transformer(summary_dim):
    from bayesflow.networks import TimeSeriesTransformer

    return TimeSeriesTransformer(summary_dim=summary_dim)


@pytest.fixture(scope="function")
def fusion_transformer(summary_dim):
    from bayesflow.networks import FusionTransformer

    return FusionTransformer(summary_dim=summary_dim)


@pytest.fixture(scope="function")
def set_transformer(summary_dim):
    from bayesflow.networks import SetTransformer

    return SetTransformer(summary_dim=summary_dim)


@pytest.fixture(scope="function")
def deep_set(summary_dim):
    from bayesflow.networks import DeepSet

    return DeepSet(summary_dim=summary_dim)


@pytest.fixture(
    params=[
        None,
        "time_series_network",
        "time_series_transformer",
        "fusion_transformer",
        "set_transformer",
        "deep_set",
    ],
    scope="function",
)
def summary_network(request, summary_dim):
    if request.param is None:
        return None
    return request.getfixturevalue(request.param)



from functools import singledispatch
import keras
import numpy as np

from bayesflow.types import Tensor
from bayesflow.types.tensor import BackendTensor


@singledispatch
def find_cost(arg, *args, **kwargs):
    raise TypeError(f"Cannot infer cost matrix from {arg!r}.")


@find_cost.register
def _(name: str, x1: Tensor, x2: Tensor, numpy: bool = False, **kwargs):
    match name.lower():
        case "euclidean":
            if numpy:
                cost = x1[:, None] - x2[None, :]
                cost = np.reshape(cost, (cost.shape[0], cost.shape[1], -1))
                cost = np.linalg.norm(cost, ord=2, axis=-1)
            else:
                cost = x1[:, None] - x2[None, :]
                cost = keras.ops.reshape(cost, (keras.ops.shape(cost)[0], keras.ops.shape(cost)[1], -1))
                cost = keras.ops.norm(cost, ord=2, axis=-1)
        case other:
            raise ValueError(f"Unsupported cost name: '{other}'.")

    return cost


@find_cost.register(BackendTensor)
def _(cost: Tensor, *args, **kwargs):
    return cost


@find_cost.register(np.ndarray)
def _(cost: np.ndarray, *args, **kwargs):
    return cost



from collections.abc import Mapping, Sequence

import keras
import numpy as np

from bayesflow.adapters import Adapter
from bayesflow.datasets import OnlineDataset
from bayesflow.networks import SummaryNetwork
from bayesflow.simulators import ModelComparisonSimulator, Simulator
from bayesflow.types import Shape, Tensor
from bayesflow.utils import filter_kwargs, logging
from bayesflow.utils.serialization import serialize, deserialize, serializable

from .approximator import Approximator


@serializable
class ModelComparisonApproximator(Approximator):
    """
    Defines an approximator for model (simulator) comparison, where the (discrete) posterior model probabilities are
    learned with a classifier.

    Parameters
    ----------
    adapter: bf.adapters.Adapter
        Adapter for data pre-processing.
    num_models: int
        Number of models (simulators) that the approximator will compare
    classifier_network: keras.Layer
        The network backbone (e.g, an MLP) that is used for model classification.
        The input of the classifier network is created by concatenating `classifier_variables`
        and (optional) output of the summary_network.
    summary_network: bf.networks.SummaryNetwork, optional
        The summary network used for data summarization (default is None).
        The input of the summary network is `summary_variables`.
    """

    SAMPLE_KEYS = ["summary_variables", "classifier_conditions"]

    def __init__(
        self,
        *,
        num_models: int,
        classifier_network: keras.Layer,
        adapter: Adapter,
        summary_network: SummaryNetwork = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.classifier_network = classifier_network
        self.adapter = adapter
        self.summary_network = summary_network
        self.num_models = num_models
        self.logits_projector = keras.layers.Dense(num_models)

    def build(self, data_shapes: Mapping[str, Shape]):
        data = {key: keras.ops.zeros(value) for key, value in data_shapes.items()}
        self.compute_metrics(**data, stage="training")

    @classmethod
    def build_adapter(
        cls,
        num_models: int,
        classifier_conditions: Sequence[str] = None,
        summary_variables: Sequence[str] = None,
        model_index_name: str = "model_indices",
    ):
        if classifier_conditions is None and summary_variables is None:
            raise ValueError("At least one of `classifier_variables` or `summary_variables` must be provided.")

        adapter = Adapter().to_array().convert_dtype("float64", "float32")

        if classifier_conditions is not None:
            adapter = adapter.concatenate(classifier_conditions, into="classifier_conditions")

        if summary_variables is not None:
            adapter = adapter.as_set(summary_variables).concatenate(summary_variables, into="summary_variables")

        adapter = (
            adapter.rename(model_index_name, "model_indices")
            .keep(["classifier_conditions", "summary_variables", "model_indices"])
            .standardize(exclude="model_indices")
            .one_hot("model_indices", num_models)
        )

        return adapter

    @classmethod
    def build_dataset(
        cls,
        *,
        dataset: keras.utils.PyDataset = None,
        simulator: ModelComparisonSimulator = None,
        simulators: Sequence[Simulator] = None,
        **kwargs,
    ) -> OnlineDataset:
        if sum(arg is not None for arg in (dataset, simulator, simulators)) != 1:
            raise ValueError("Exactly one of dataset, simulator, or simulators must be provided.")

        if simulators is not None:
            simulator = ModelComparisonSimulator(simulators)

        return super().build_dataset(dataset=dataset, simulator=simulator, **kwargs)

    def compile(
        self,
        *args,
        classifier_metrics: Sequence[keras.Metric] = None,
        summary_metrics: Sequence[keras.Metric] = None,
        **kwargs,
    ):
        if classifier_metrics:
            self.classifier_network._metrics = classifier_metrics

        if summary_metrics:
            if self.summary_network is None:
                logging.warning("Ignoring summary metrics because there is no summary network.")
            else:
                self.summary_network._metrics = summary_metrics

        return super().compile(*args, **kwargs)

    def compile_from_config(self, config):
        self.compile(**deserialize(config))
        if hasattr(self, "optimizer") and self.built:
            # Create optimizer variables.
            self.optimizer.build(self.trainable_variables)

    def compute_metrics(
        self,
        *,
        classifier_conditions: Tensor = None,
        model_indices: Tensor,
        summary_variables: Tensor = None,
        stage: str = "training",
    ) -> dict[str, Tensor]:
        if self.summary_network is None:
            summary_metrics = {}
        else:
            summary_metrics = self.summary_network.compute_metrics(summary_variables, stage=stage)
            summary_outputs = summary_metrics.pop("outputs")

            if classifier_conditions is None:
                classifier_conditions = summary_outputs
            else:
                classifier_conditions = keras.ops.concatenate([classifier_conditions, summary_outputs], axis=-1)

        # we could move this into its own class
        logits = self.classifier_network(classifier_conditions)
        logits = self.logits_projector(logits)

        cross_entropy = keras.losses.categorical_crossentropy(model_indices, logits, from_logits=True)
        cross_entropy = keras.ops.mean(cross_entropy)

        classifier_metrics = {"loss": cross_entropy}

        if stage != "training" and any(self.classifier_network.metrics):
            # compute sample-based metrics
            predictions = keras.ops.argmax(logits, axis=-1)
            classifier_metrics |= {
                metric.name: metric(model_indices, predictions) for metric in self.classifier_network.metrics
            }

        loss = classifier_metrics.get("loss", keras.ops.zeros(())) + summary_metrics.get("loss", keras.ops.zeros(()))

        classifier_metrics = {f"{key}/classifier_{key}": value for key, value in classifier_metrics.items()}
        summary_metrics = {f"{key}/summary_{key}": value for key, value in summary_metrics.items()}

        metrics = {"loss": loss} | classifier_metrics | summary_metrics

        return metrics

    def fit(
        self,
        *,
        adapter: Adapter = "auto",
        dataset: keras.utils.PyDataset = None,
        simulator: ModelComparisonSimulator = None,
        simulators: Sequence[Simulator] = None,
        **kwargs,
    ):
        """
        Trains the approximator on the provided dataset or on-demand generated from the given (multi-model) simulator.
        If `dataset` is not provided, a dataset is built from the `simulator`.
        If `simulator` is not provided, it will be build from a list of `simulators`.
        If the model has not been built, it will be built using a batch from the dataset.

        Parameters
        ----------
        dataset : keras.utils.PyDataset, optional
            A dataset containing simulations for training. If provided, `simulator` must be None.
        simulator : ModelComparisonSimulator, optional
            A simulator used to generate a dataset. If provided, `dataset` must be None.
        simulators: Sequence[Simulator], optional
            A list of simulators (one simulator per model). If provided, `dataset` must be None.
        **kwargs : dict
            Additional keyword arguments passed to `keras.Model.fit()`, including (see also `build_dataset`):

            batch_size : int or None, default='auto'
                Number of samples per gradient update. Do not specify if `dataset` is provided as a
                `keras.utils.PyDataset`, `tf.data.Dataset`, `torch.utils.data.DataLoader`, or a generator function.
            epochs : int, default=1
                Number of epochs to train the model.
            verbose : {"auto", 0, 1, 2}, default="auto"
                Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line per epoch.
            callbacks : list of keras.callbacks.Callback, optional
                List of callbacks to apply during training.
            validation_split : float, optional
                Fraction of training data to use for validation (only supported if `dataset` consists of NumPy arrays
                or tensors).
            validation_data : tuple or dataset, optional
                Data for validation, overriding `validation_split`.
            shuffle : bool, default=True
                Whether to shuffle the training data before each epoch (ignored for dataset generators).
            initial_epoch : int, default=0
                Epoch at which to start training (useful for resuming training).
            steps_per_epoch : int or None, optional
                Number of steps (batches) before declaring an epoch finished.
            validation_steps : int or None, optional
                Number of validation steps per validation epoch.
            validation_batch_size : int or None, optional
                Number of samples per validation batch (defaults to `batch_size`).
            validation_freq : int, default=1
                Specifies how many training epochs to run before performing validation.

        Returns
        -------
        keras.callbacks.History
            A history object containing the training loss and metrics values.

        Raises
        ------
        ValueError
            If both `dataset` and `simulator` or `simulators` are provided or neither is provided.
        """
        if dataset is not None:
            if simulator is not None or simulators is not None:
                raise ValueError(
                    "Received conflicting arguments. Please provide either a dataset or a simulator, but not both."
                )

            return super().fit(dataset=dataset, **kwargs)

        if adapter == "auto":
            logging.info("Building automatic data adapter.")
            adapter = self.build_adapter(num_models=self.num_models, **filter_kwargs(kwargs, self.build_adapter))

        if simulator is not None:
            return super().fit(simulator=simulator, adapter=adapter, **kwargs)

        logging.info(f"Building model comparison simulator from {len(simulators)} simulators.")

        simulator = ModelComparisonSimulator(simulators=simulators)

        return super().fit(simulator=simulator, adapter=adapter, **kwargs)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()

        config = {
            "num_models": self.num_models,
            "adapter": self.adapter,
            "classifier_network": self.classifier_network,
            "summary_network": self.summary_network,
        }

        return base_config | serialize(config)

    def get_compile_config(self):
        base_config = super().get_compile_config() or {}

        config = {
            "classifier_metrics": self.classifier_network._metrics,
            "summary_metrics": self.summary_network._metrics if self.summary_network is not None else None,
        }

        return base_config | serialize(config)

    def predict(
        self,
        *,
        conditions: Mapping[str, np.ndarray],
        logits: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """
        Predicts posterior model probabilities given input conditions. The `conditions` dictionary is preprocessed
        using the `adapter`. The output is converted to NumPy array after inference.

        Parameters
        ----------
        conditions : Mapping[str, np.ndarray]
            Dictionary of conditioning variables as NumPy arrays.
        logits: bool, default=False
            Should the posterior model probabilities be on the (unconstrained) logit space?
            If `False`, the output is a unit simplex instead.
        **kwargs : dict
            Additional keyword arguments for the adapter and classification process.

        Returns
        -------
        np.ndarray
            Predicted posterior model probabilities given `conditions`.
        """

        # Apply adapter transforms to raw simulated / real quantities
        conditions = self.adapter(conditions, strict=False, stage="inference", **kwargs)

        # Ensure only keys relevant for sampling are present in the conditions dictionary
        conditions = {k: v for k, v in conditions.items() if k in ModelComparisonApproximator.SAMPLE_KEYS}

        conditions = keras.tree.map_structure(keras.ops.convert_to_tensor, conditions)

        output = self._predict(**conditions, **kwargs)

        if not logits:
            output = keras.ops.softmax(output)

        output = keras.ops.convert_to_numpy(output)

        return output

    def _predict(self, classifier_conditions: Tensor = None, summary_variables: Tensor = None, **kwargs) -> Tensor:
        if self.summary_network is None:
            if summary_variables is not None:
                raise ValueError("Cannot use summary variables without a summary network.")
        else:
            if summary_variables is None:
                raise ValueError("Summary variables are required when a summary network is present")

            summary_outputs = self.summary_network(
                summary_variables, **filter_kwargs(kwargs, self.summary_network.call)
            )

            if classifier_conditions is None:
                classifier_conditions = summary_outputs
            else:
                classifier_conditions = keras.ops.concatenate([classifier_conditions, summary_outputs], axis=1)

        output = self.classifier_network(classifier_conditions)
        output = self.logits_projector(output)

        return output



import keras
import numpy as np


def allclose(x1, x2, rtol=1e-5, atol=1e-5):
    return keras.ops.all(keras.ops.isclose(x1, x2, rtol, atol))


def assert_allclose(x1, x2, rtol=1e-5, atol=1e-8, msg=""):
    x1 = keras.ops.convert_to_numpy(x1)
    x2 = keras.ops.convert_to_numpy(x2)

    assert x1.shape == x2.shape, "Input shapes do not match."

    mse = np.mean(np.square(x1 - x2)).item()
    largest_deviation = np.max(np.abs(x1 - x2)).item()
    largest_deviation_index = np.unravel_index(np.argmax(np.abs(x1 - x2)), x1.shape)
    largest_deviation_value1 = x1[largest_deviation_index].item()
    largest_deviation_value2 = x2[largest_deviation_index].item()

    if msg:
        msg = f"{msg}\n"
    else:
        msg = "Inputs significantly differ:\n"

    msg += "Largest Deviation:\n"
    msg += f"|{largest_deviation_value1:.02e} - {largest_deviation_value2:.02e}| = {largest_deviation:.02e}\n"
    msg += "\n"
    msg += "MSE:\n"
    msg += f"{mse:.02e}"

    assert allclose(x1, x2, rtol, atol), msg


def max_mean_discrepancy(x, y):
    # Computes the Max Mean Discrepancy between samples of two distributions
    xx = keras.ops.matmul(x, keras.ops.transpose(x))
    yy = keras.ops.matmul(y, keras.ops.transpose(y))
    zz = keras.ops.matmul(x, keras.ops.transpose(y))

    rx = keras.ops.broadcast_to(keras.ops.expand_dims(keras.ops.diag(xx), 0), xx.shape)
    ry = keras.ops.broadcast_to(keras.ops.expand_dims(keras.ops.diag(yy), 0), yy.shape)

    dxx = keras.ops.transpose(rx) + rx - 2.0 * xx
    dyy = keras.ops.transpose(ry) + ry - 2.0 * yy
    dxy = keras.ops.transpose(rx) + ry - 2.0 * zz

    XX = keras.ops.zeros(xx.shape)
    YY = keras.ops.zeros(yy.shape)
    XY = keras.ops.zeros(zz.shape)

    # RBF scaling
    bandwidth = [10, 15, 20, 50]
    for a in bandwidth:
        XX += keras.ops.exp(-0.5 * dxx / a)
        YY += keras.ops.exp(-0.5 * dyy / a)
        XY += keras.ops.exp(-0.5 * dxy / a)

    return keras.ops.mean(XX + YY - 2.0 * XY)



from collections.abc import Callable
import numpy as np
from keras.saving import (
    deserialize_keras_object as deserialize,
    register_keras_serializable as serializable,
    serialize_keras_object as serialize,
    get_registered_name,
    get_registered_object,
)
from .elementwise_transform import ElementwiseTransform
from ...utils import filter_kwargs
import inspect


@serializable(package="bayesflow.adapters")
class SerializableCustomTransform(ElementwiseTransform):
    """
    Transforms a parameter using a pair of registered serializable forward and inverse functions.

    Parameters
    ----------
    forward : function, no lambda
        Registered serializable function to transform the data in the forward pass.
        For the adapter to be serializable, this function has to be serializable
        as well (see Notes). Therefore, only proper functions and no lambda
        functions can be used here.
    inverse : function, no lambda
        Function to transform the data in the inverse pass.
        For the adapter to be serializable, this function has to be serializable
        as well (see Notes). Therefore, only proper functions and no lambda
        functions can be used here.

    Raises
    ------
    ValueError
        When the provided functions are not registered serializable functions.

    Notes
    -----
    Important: The forward and inverse functions have to be registered with Keras.
    To do so, use the `@keras.saving.register_keras_serializable` decorator.
    They must also be registered (and identical) when loading the adapter
    at a later point in time.

    """

    def __init__(
        self,
        *,
        forward: Callable[[np.ndarray, ...], np.ndarray],
        inverse: Callable[[np.ndarray, ...], np.ndarray],
    ):
        super().__init__()

        self._check_serializable(forward, label="forward")
        self._check_serializable(inverse, label="inverse")
        self._forward = forward
        self._inverse = inverse

    @classmethod
    def _check_serializable(cls, function, label=""):
        GENERAL_EXAMPLE_CODE = (
            "The example code below shows the structure of a correctly decorated function:\n\n"
            "```\n"
            "import keras\n\n"
            "@keras.saving.register_keras_serializable('custom')\n"
            f"def my_{label}(...):\n"
            "    [your code goes here...]\n"
            "```\n"
        )
        if function is None:
            raise TypeError(
                f"'{label}' must be a registered serializable function, was 'NoneType'.\n{GENERAL_EXAMPLE_CODE}"
            )
        registered_name = get_registered_name(function)
        # check if function is a lambda function
        if registered_name == "<lambda>":
            raise ValueError(
                f"The provided function for '{label}' is a lambda function, "
                "which cannot be serialized. "
                "Please provide a registered serializable function by using the "
                "@keras.saving.register_keras_serializable decorator."
                f"\n{GENERAL_EXAMPLE_CODE}"
            )
        if inspect.ismethod(function):
            raise ValueError(
                f"The provided value for '{label}' is a method, not a function. "
                "Methods cannot be serialized separately from their classes. "
                "Please provide a registered serializable function instead by "
                "moving the functionality to a function (i.e., outside of the class) and "
                "using the @keras.saving.register_keras_serializable decorator."
                f"\n{GENERAL_EXAMPLE_CODE}"
            )
        registered_object_for_name = get_registered_object(registered_name)
        if registered_object_for_name is None:
            try:
                source_max_lines = 5
                function_source_code = inspect.getsource(function).split("\n")
                if len(function_source_code) > source_max_lines:
                    function_source_code = function_source_code[:source_max_lines] + ["    [...]"]

                example_code = "For your provided function, this would look like this:\n\n"
                example_code += "\n".join(
                    ["```", "import keras\n", "@keras.saving.register_keras_serializable('custom')"]
                    + function_source_code
                    + ["```"]
                )
            except OSError:
                example_code = GENERAL_EXAMPLE_CODE
            raise ValueError(
                f"The provided function for '{label}' is not registered with Keras.\n"
                "Please register the function using the "
                "@keras.saving.register_keras_serializable decorator.\n"
                f"{example_code}"
            )
        if registered_object_for_name is not function:
            raise ValueError(
                f"The provided function for '{label}' does not match the function "
                f"registered under its name '{registered_name}'. "
                f"(registered function: {registered_object_for_name}, provided function: {function}). "
            )

    @classmethod
    def from_config(cls, config: dict, custom_objects=None) -> "SerializableCustomTransform":
        if get_registered_object(config["forward"]["config"], custom_objects) is None:
            provided_function_msg = ""
            if config["_forward_source_code"]:
                provided_function_msg = (
                    f"\nThe originally provided function was:\n\n```\n{config['_forward_source_code']}\n```"
                )
            raise TypeError(
                "\n\nPLEASE READ HERE:\n"
                "-----------------\n"
                "The forward function that was provided as `forward` "
                "is not registered with Keras, making deserialization impossible. "
                f"Please ensure that it is registered as '{config['forward']['config']}' and identical to the original "
                "function before loading your model."
                f"{provided_function_msg}"
            )
        if get_registered_object(config["inverse"]["config"], custom_objects) is None:
            provided_function_msg = ""
            if config["_inverse_source_code"]:
                provided_function_msg = (
                    f"\nThe originally provided function was:\n\n```\n{config['_inverse_source_code']}\n```"
                )
            raise TypeError(
                "\n\nPLEASE READ HERE:\n"
                "-----------------\n"
                "The inverse function that was provided as `inverse` "
                "is not registered with Keras, making deserialization impossible. "
                f"Please ensure that it is registered as '{config['inverse']['config']}' and identical to the original "
                "function before loading your model."
                f"{provided_function_msg}"
            )
        forward = deserialize(config["forward"], custom_objects)
        inverse = deserialize(config["inverse"], custom_objects)
        return cls(
            forward=forward,
            inverse=inverse,
        )

    def get_config(self) -> dict:
        forward_source_code = inverse_source_code = None
        try:
            forward_source_code = inspect.getsource(self._forward)
            inverse_source_code = inspect.getsource(self._inverse)
        except OSError:
            pass
        return {
            "forward": serialize(self._forward),
            "inverse": serialize(self._inverse),
            "_forward_source_code": forward_source_code,
            "_inverse_source_code": inverse_source_code,
        }

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        # filter kwargs so that other transform args like batch_size, strict, ... are not passed through
        kwargs = filter_kwargs(kwargs, self._forward)
        return self._forward(data, **kwargs)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        kwargs = filter_kwargs(kwargs, self._inverse)
        return self._inverse(data, **kwargs)



import bayesflow as bf
import pytest


def num_variables(x: dict):
    return sum(arr.shape[-1] for arr in x.values())


def test_metric_calibration_error(random_estimates, random_targets, var_names):
    # basic functionality: automatic variable names
    out = bf.diagnostics.metrics.calibration_error(random_estimates, random_targets)
    assert list(out.keys()) == ["values", "metric_name", "variable_names"]
    assert out["values"].shape == (num_variables(random_estimates),)
    assert out["metric_name"] == "Calibration Error"
    assert out["variable_names"] == ["beta_0", "beta_1", "sigma"]

    # user specified variable names
    out = bf.diagnostics.metrics.calibration_error(
        estimates=random_estimates,
        targets=random_targets,
        variable_names=var_names,
    )
    assert out["variable_names"] == var_names

    # user-specifed keys and scalar variable
    out = bf.diagnostics.metrics.calibration_error(
        estimates=random_estimates,
        targets=random_targets,
        variable_keys="sigma",
    )
    assert out["values"].shape == (random_estimates["sigma"].shape[-1],)
    assert out["variable_names"] == ["sigma"]


def test_posterior_contraction(random_estimates, random_targets):
    # basic functionality: automatic variable names
    out = bf.diagnostics.metrics.posterior_contraction(random_estimates, random_targets)
    assert list(out.keys()) == ["values", "metric_name", "variable_names"]
    assert out["values"].shape == (num_variables(random_estimates),)
    assert out["metric_name"] == "Posterior Contraction"
    assert out["variable_names"] == ["beta_0", "beta_1", "sigma"]


def test_root_mean_squared_error(random_estimates, random_targets):
    # basic functionality: automatic variable names
    out = bf.diagnostics.metrics.root_mean_squared_error(random_estimates, random_targets)
    assert list(out.keys()) == ["values", "metric_name", "variable_names"]
    assert out["values"].shape == (num_variables(random_estimates),)
    assert out["metric_name"] == "NRMSE"
    assert out["variable_names"] == ["beta_0", "beta_1", "sigma"]


def test_classifier_two_sample_test(random_samples_a, random_samples_b):
    metric = bf.diagnostics.metrics.classifier_two_sample_test(estimates=random_samples_a, targets=random_samples_a)
    assert 0.55 > metric > 0.45

    metric = bf.diagnostics.metrics.classifier_two_sample_test(estimates=random_samples_a, targets=random_samples_b)
    assert metric > 0.55


def test_expected_calibration_error(pred_models, true_models, model_names):
    out = bf.diagnostics.metrics.expected_calibration_error(pred_models, true_models, model_names=model_names)
    assert list(out.keys()) == ["values", "metric_name", "model_names"]
    assert out["values"].shape == (pred_models.shape[-1],)
    assert out["metric_name"] == "Expected Calibration Error"
    assert out["model_names"] == [r"$\mathcal{M}_0$", r"$\mathcal{M}_1$", r"$\mathcal{M}_2$"]

    # returns probs?
    out = bf.diagnostics.metrics.expected_calibration_error(pred_models, true_models, return_probs=True)
    assert list(out.keys()) == ["values", "metric_name", "model_names", "probs_true", "probs_pred"]
    assert len(out["probs_true"]) == pred_models.shape[-1]
    assert len(out["probs_pred"]) == pred_models.shape[-1]
    # default: auto model names
    assert out["model_names"] == ["M_0", "M_1", "M_2"]

    # handles incorrect input?
    with pytest.raises(Exception):
        out = bf.diagnostics.metrics.expected_calibration_error(pred_models, true_models, model_names=["a"])

    with pytest.raises(Exception):
        out = bf.diagnostics.metrics.expected_calibration_error(pred_models, true_models.transpose)



def docstring(app, what, name, obj, options, lines):
    """Adapt autodoc docstring.

    Inherited Keras docstrings do not follow the numpy docstring format.
    Most noticably, they use Markdown code fences. To improve the situation
    somewhat, this functions aims to convert fenced code blocks to RST
    code blocks.
    """
    updated_lines = []
    prefix = ""
    for line in lines:
        if line.count("```") == 1:
            if prefix == "":
                prefix = "    "
                updated_lines[-1] = updated_lines[-1] + "::"
            else:
                prefix = ""
            updated_lines.append("\n")
        else:
            updated_lines.append(prefix + line)
    if prefix != "":
        raise ValueError(
            f"Uneven number of code fences in docstring:\nwhat='{what}', name='{name}'.\n" + "\n".join(lines)
        )
    # overwrite lines list
    lines.clear()
    lines += updated_lines


def setup(app):
    app.connect("autodoc-process-docstring", docstring)



from collections.abc import Mapping, Sequence, Callable

import numpy as np

import keras

from bayesflow.adapters import Adapter
from bayesflow.networks import InferenceNetwork, SummaryNetwork
from bayesflow.types import Tensor
from bayesflow.utils import filter_kwargs, logging, split_arrays, squeeze_inner_estimates_dict
from bayesflow.utils.serialization import serialize, deserialize, serializable

from .approximator import Approximator


@serializable
class ContinuousApproximator(Approximator):
    """
    Defines a workflow for performing fast posterior or likelihood inference.
    The distribution is approximated with an inference network and an optional summary network.

    Parameters
    ----------
    adapter : bayesflow.adapters.Adapter
        Adapter for data processing. You can use :py:meth:`build_adapter`
        to create it.
    inference_network : InferenceNetwork
        The inference network used for posterior or likelihood approximation.
    summary_network : SummaryNetwork, optional
        The summary network used for data summarization (default is None).
    **kwargs : dict, optional
        Additional arguments passed to the :py:class:`bayesflow.approximators.Approximator` class.
    """

    SAMPLE_KEYS = ["summary_variables", "inference_conditions"]

    def __init__(
        self,
        *,
        adapter: Adapter,
        inference_network: InferenceNetwork,
        summary_network: SummaryNetwork = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.adapter = adapter
        self.inference_network = inference_network
        self.summary_network = summary_network

    @classmethod
    def build_adapter(
        cls,
        inference_variables: Sequence[str],
        inference_conditions: Sequence[str] = None,
        summary_variables: Sequence[str] = None,
        standardize: bool = True,
        sample_weight: str = None,
    ) -> Adapter:
        """Create an :py:class:`~bayesflow.adapters.Adapter` suited for the approximator.

        Parameters
        ----------
        inference_variables : Sequence of str
            Names of the inference variables in the data
        inference_conditions : Sequence of str, optional
            Names of the inference conditions in the data
        summary_variables : Sequence of str, optional
            Names of the summary variables in the data
        standardize : bool, optional
            Decide whether to standardize all variables, default is True
        sample_weight : str, optional
            Name of the sample weights
        """

        adapter = Adapter()
        adapter.to_array()
        adapter.convert_dtype("float64", "float32")
        adapter.concatenate(inference_variables, into="inference_variables")

        if inference_conditions is not None:
            adapter.concatenate(inference_conditions, into="inference_conditions")

        if summary_variables is not None:
            adapter.as_set(summary_variables)
            adapter.concatenate(summary_variables, into="summary_variables")

        if sample_weight is not None:
            adapter = adapter.rename(sample_weight, "sample_weight")

        adapter.keep(["inference_variables", "inference_conditions", "summary_variables", "sample_weight"])

        if standardize:
            adapter.standardize(exclude="sample_weight")

        return adapter

    def compile(
        self,
        *args,
        inference_metrics: Sequence[keras.Metric] = None,
        summary_metrics: Sequence[keras.Metric] = None,
        **kwargs,
    ):
        if inference_metrics:
            self.inference_network._metrics = inference_metrics

        if summary_metrics:
            if self.summary_network is None:
                logging.warning("Ignoring summary metrics because there is no summary network.")
            else:
                self.summary_network._metrics = summary_metrics

        return super().compile(*args, **kwargs)

    def compile_from_config(self, config):
        self.compile(**deserialize(config))
        if hasattr(self, "optimizer") and self.built:
            # Create optimizer variables.
            self.optimizer.build(self.trainable_variables)

    def compute_metrics(
        self,
        inference_variables: Tensor,
        inference_conditions: Tensor = None,
        summary_variables: Tensor = None,
        sample_weight: Tensor = None,
        stage: str = "training",
    ) -> dict[str, Tensor]:
        if self.summary_network is None:
            if summary_variables is not None:
                raise ValueError("Cannot compute summary metrics without a summary network.")

            summary_metrics = {}
        else:
            if summary_variables is None:
                raise ValueError("Summary variables are required when a summary network is present.")

            summary_metrics = self.summary_network.compute_metrics(summary_variables, stage=stage)
            summary_outputs = summary_metrics.pop("outputs")

            # append summary outputs to inference conditions
            if inference_conditions is None:
                inference_conditions = summary_outputs
            else:
                inference_conditions = keras.ops.concatenate([inference_conditions, summary_outputs], axis=-1)

        # Force a conversion to Tensor
        inference_variables = keras.tree.map_structure(keras.ops.convert_to_tensor, inference_variables)
        inference_metrics = self.inference_network.compute_metrics(
            inference_variables, conditions=inference_conditions, sample_weight=sample_weight, stage=stage
        )

        loss = inference_metrics.get("loss", keras.ops.zeros(())) + summary_metrics.get("loss", keras.ops.zeros(()))

        inference_metrics = {f"{key}/inference_{key}": value for key, value in inference_metrics.items()}
        summary_metrics = {f"{key}/summary_{key}": value for key, value in summary_metrics.items()}

        metrics = {"loss": loss} | inference_metrics | summary_metrics

        return metrics

    def fit(self, *args, **kwargs):
        """
        Trains the approximator on the provided dataset or on-demand data generated from the given simulator.
        If `dataset` is not provided, a dataset is built from the `simulator`.
        If the model has not been built, it will be built using a batch from the dataset.

        Parameters
        ----------
        dataset : keras.utils.PyDataset, optional
            A dataset containing simulations for training. If provided, `simulator` must be None.
        simulator : Simulator, optional
            A simulator used to generate a dataset. If provided, `dataset` must be None.
        **kwargs
            Additional keyword arguments passed to `keras.Model.fit()`, including (see also `build_dataset`):

            batch_size : int or None, default='auto'
                Number of samples per gradient update. Do not specify if `dataset` is provided as a
                `keras.utils.PyDataset`, `tf.data.Dataset`, `torch.utils.data.DataLoader`, or a generator function.
            epochs : int, default=1
                Number of epochs to train the model.
            verbose : {"auto", 0, 1, 2}, default="auto"
                Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line per epoch.
            callbacks : list of keras.callbacks.Callback, optional
                List of callbacks to apply during training.
            validation_split : float, optional
                Fraction of training data to use for validation (only supported if `dataset` consists of NumPy arrays
                or tensors).
            validation_data : tuple or dataset, optional
                Data for validation, overriding `validation_split`.
            shuffle : bool, default=True
                Whether to shuffle the training data before each epoch (ignored for dataset generators).
            initial_epoch : int, default=0
                Epoch at which to start training (useful for resuming training).
            steps_per_epoch : int or None, optional
                Number of steps (batches) before declaring an epoch finished.
            validation_steps : int or None, optional
                Number of validation steps per validation epoch.
            validation_batch_size : int or None, optional
                Number of samples per validation batch (defaults to `batch_size`).
            validation_freq : int, default=1
                Specifies how many training epochs to run before performing validation.

        Returns
        -------
        keras.callbacks.History
            A history object containing the training loss and metrics values.

        Raises
        ------
        ValueError
            If both `dataset` and `simulator` are provided or neither is provided.
        """
        return super().fit(*args, **kwargs, adapter=self.adapter)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()
        config = {
            "adapter": self.adapter,
            "inference_network": self.inference_network,
            "summary_network": self.summary_network,
        }

        return base_config | serialize(config)

    def get_compile_config(self):
        base_config = super().get_compile_config() or {}

        config = {
            "inference_metrics": self.inference_network._metrics,
            "summary_metrics": self.summary_network._metrics if self.summary_network is not None else None,
        }

        return base_config | serialize(config)

    def estimate(
        self,
        conditions: Mapping[str, np.ndarray],
        split: bool = False,
        estimators: Mapping[str, Callable] = None,
        num_samples: int = 1000,
        **kwargs,
    ) -> dict[str, dict[str, np.ndarray]]:
        """
        Estimate summary statistics for variables based on given conditions.

        This function samples data using the object's ``sample`` method according to the provided
        conditions and then computes summary statistics for each variable using a set of estimator
        functions. By default, it calculates the mean, median, and selected quantiles (10th, 50th,
        and 90th percentiles). Users can also supply custom estimators that override or extend the
        default ones.

        Parameters
        ----------
        conditions : Mapping[str, np.ndarray]
            A mapping from variable names to numpy arrays representing the conditions under which
            samples should be generated.
        split : bool, optional
            If True, indicates that the data sampling process should split the samples based on an
            internal logic. The default is False.
        estimators : Mapping[str, Callable], optional
            A dictionary where keys are estimator names and values are callables. Each callable must
            accept an array and an axis parameter, and return a dictionary with the computed statistic.
            If not provided, a default set of estimators is used:
                - 'mean': Computes the mean along the specified axis.
                - 'median': Computes the median along the specified axis.
                - 'quantiles': Computes the 10th, 50th, and 90th percentiles along the specified axis,
                  then rearranges the axes for convenience.
        num_samples : int, optional
            The number of samples to generate for each variable. The default is 1000.
        **kwargs
            Additional keyword arguments passed to the ``sample`` method.

        Returns
        -------
        dict[str, dict[str, np.ndarray]]
            A nested dictionary where the outer keys correspond to variable names and the inner keys
            correspond to estimator names. Each inner dictionary contains the computed statistic(s) for
            the variable, potentially with reduced nesting via ``squeeze_inner_estimates_dict``.
        """

        estimators = estimators or {}
        estimators = (
            dict(
                mean=lambda x, axis: dict(value=np.mean(x, keepdims=True, axis=axis)),
                median=lambda x, axis: dict(value=np.median(x, keepdims=True, axis=axis)),
                quantiles=lambda x, axis: dict(value=np.moveaxis(np.quantile(x, q=[0.1, 0.5, 0.9], axis=axis), 0, 1)),
            )
            | estimators
        )

        samples = self.sample(num_samples=num_samples, conditions=conditions, split=split, **kwargs)

        estimates = {
            variable_name: {
                estimator_name: func(samples[variable_name], axis=1) for estimator_name, func in estimators.items()
            }
            for variable_name in samples.keys()
        }

        # remove unnecessary nesting
        estimates = {
            variable_name: {
                outer_key: squeeze_inner_estimates_dict(estimates[variable_name][outer_key])
                for outer_key in estimates[variable_name].keys()
            }
            for variable_name in estimates.keys()
        }

        return estimates

    def sample(
        self,
        *,
        num_samples: int,
        conditions: Mapping[str, np.ndarray],
        split: bool = False,
        **kwargs,
    ) -> dict[str, np.ndarray]:
        """
        Generates samples from the approximator given input conditions. The `conditions` dictionary is preprocessed
        using the `adapter`. Samples are converted to NumPy arrays after inference.

        Parameters
        ----------
        num_samples : int
            Number of samples to generate.
        conditions : dict[str, np.ndarray]
            Dictionary of conditioning variables as NumPy arrays.
        split : bool, default=False
            Whether to split the output arrays along the last axis and return one column vector per target variable
            samples.
        **kwargs : dict
            Additional keyword arguments for the adapter and sampling process.

        Returns
        -------
        dict[str, np.ndarray]
            Dictionary containing generated samples with the same keys as `conditions`.
        """

        # Apply adapter transforms to raw simulated / real quantities
        conditions = self.adapter(conditions, strict=False, stage="inference", **kwargs)

        # Ensure only keys relevant for sampling are present in the conditions dictionary
        conditions = {k: v for k, v in conditions.items() if k in ContinuousApproximator.SAMPLE_KEYS}

        conditions = keras.tree.map_structure(keras.ops.convert_to_tensor, conditions)
        conditions = {"inference_variables": self._sample(num_samples=num_samples, **conditions, **kwargs)}
        conditions = keras.tree.map_structure(keras.ops.convert_to_numpy, conditions)

        # Back-transform quantities and samples
        conditions = self.adapter(conditions, inverse=True, strict=False, **kwargs)

        if split:
            conditions = split_arrays(conditions, axis=-1)
        return conditions

    def _sample(
        self,
        num_samples: int,
        inference_conditions: Tensor = None,
        summary_variables: Tensor = None,
        **kwargs,
    ) -> Tensor:
        if self.summary_network is None:
            if summary_variables is not None:
                raise ValueError("Cannot use summary variables without a summary network.")
        else:
            if summary_variables is None:
                raise ValueError("Summary variables are required when a summary network is present.")

            summary_outputs = self.summary_network(
                summary_variables, **filter_kwargs(kwargs, self.summary_network.call)
            )

            if inference_conditions is None:
                inference_conditions = summary_outputs
            else:
                inference_conditions = keras.ops.concatenate([inference_conditions, summary_outputs], axis=1)

        if inference_conditions is not None:
            # conditions must always have shape (batch_size, dims)
            batch_size = keras.ops.shape(inference_conditions)[0]
            inference_conditions = keras.ops.expand_dims(inference_conditions, axis=1)
            inference_conditions = keras.ops.broadcast_to(
                inference_conditions, (batch_size, num_samples, *keras.ops.shape(inference_conditions)[2:])
            )
            batch_shape = (batch_size, num_samples)
        else:
            batch_shape = (num_samples,)

        return self.inference_network.sample(
            batch_shape,
            conditions=inference_conditions,
            **filter_kwargs(kwargs, self.inference_network.sample),
        )

    def log_prob(self, data: Mapping[str, np.ndarray], **kwargs) -> np.ndarray | dict[str, np.ndarray]:
        """
        Computes the log-probability of given data under the model. The `data` dictionary is preprocessed using the
        `adapter`. Log-probabilities are returned as NumPy arrays.

        Parameters
        ----------
        data : Mapping[str, np.ndarray]
            Dictionary of observed data as NumPy arrays.
        **kwargs : dict
            Additional keyword arguments for the adapter and log-probability computation.

        Returns
        -------
        np.ndarray
            Log-probabilities of the distribution `p(inference_variables | inference_conditions, h(summary_conditions))`
        """
        data = self.adapter(data, strict=False, stage="inference", **kwargs)
        data = keras.tree.map_structure(keras.ops.convert_to_tensor, data)
        log_prob = self._log_prob(**data, **kwargs)
        log_prob = keras.tree.map_structure(keras.ops.convert_to_numpy, log_prob)

        return log_prob

    def _log_prob(
        self,
        inference_variables: Tensor,
        inference_conditions: Tensor = None,
        summary_variables: Tensor = None,
        **kwargs,
    ) -> Tensor | dict[str, Tensor]:
        if self.summary_network is None:
            if summary_variables is not None:
                raise ValueError("Cannot use summary variables without a summary network.")
        else:
            if summary_variables is None:
                raise ValueError("Summary variables are required when a summary network is present.")

            summary_outputs = self.summary_network(
                summary_variables, **filter_kwargs(kwargs, self.summary_network.call)
            )

            if inference_conditions is None:
                inference_conditions = summary_outputs
            else:
                inference_conditions = keras.ops.concatenate([inference_conditions, summary_outputs], axis=-1)

        return self.inference_network.log_prob(
            inference_variables,
            conditions=inference_conditions,
            **filter_kwargs(kwargs, self.inference_network.log_prob),
        )



import keras

from bayesflow.types import Tensor
from bayesflow.utils import expand_tile

from bayesflow.utils.serialization import serializable


@serializable
class RecurrentEmbedding(keras.Layer):
    """Implements a recurrent network for flexibly embedding time vectors."""

    def __init__(self, embed_dim: int = 8, embedding: str = "lstm"):
        super().__init__()

        self.embed_dim = embed_dim
        self.embedding = embedding
        if embedding == "lstm":
            self.embedder = keras.layers.LSTM(embed_dim, return_sequences=True)
        elif embedding == "gru":
            self.embedder = keras.layers.GRU(embed_dim, return_sequences=True)
        else:
            raise ValueError(f"Unknown embedding type {embedding}. Must be in ['lstm', 'gru']")

    def call(self, x: Tensor, t: Tensor = None) -> Tensor:
        """Creates time representations and concatenates them to x.

        Parameters
        ----------
        x   : Tensor of shape (batch_size, sequence_length, dim)
            The input sequence.
        t   : Tensor of shape (batch_size, sequence_length)
            Vector of times

        Returns
        -------
        emb : Tensor
            Embedding of shape (batch_size, sequence_length, embed_dim + 1)
        """

        if t is None:
            t = keras.ops.linspace(0, keras.ops.shape(x)[1], keras.ops.shape(x)[1], dtype=x.dtype)
            t = expand_tile(t, keras.ops.shape(x)[0], axis=0)

        emb = self.embedder(t[..., None])
        return keras.ops.concatenate([x, emb], axis=-1)



from collections.abc import Sequence

import keras

from bayesflow.types import Tensor
from bayesflow.utils.serialization import serializable

from .equivariant_layer import EquivariantLayer
from .invariant_layer import InvariantLayer

from ..summary_network import SummaryNetwork


@serializable
class DeepSet(SummaryNetwork):
    """Implements a deep set encoder introduced in [1] for learning permutation-invariant representations of
    set-based data, as generated by exchangeable models.

    [1] Zaheer, M., Kottur, S., Ravanbakhsh, S., Poczos, B., Salakhutdinov, R. R., & Smola, A. J. (2017).
    Deep sets. Advances in neural information processing systems, 30.
    """

    def __init__(
        self,
        summary_dim: int = 16,
        depth: int = 2,
        inner_pooling: str = "mean",
        output_pooling: str = "mean",
        mlp_widths_equivariant: Sequence[int] = (64, 64),
        mlp_widths_invariant_inner: Sequence[int] = (64, 64),
        mlp_widths_invariant_outer: Sequence[int] = (64, 64),
        mlp_widths_invariant_last: Sequence[int] = (64, 64),
        activation: str = "gelu",
        kernel_initializer: str = "he_normal",
        dropout: int | float | None = 0.05,
        spectral_normalization: bool = False,
        **kwargs,
    ):
        """
        Initializes a fully customizable deep learning model for learning permutation-invariant representations of
        sets (i.e., exchangeable or IID data). Do not use this model for non-IID data (e.g., time series).

        Important: Prefer a SetTransformer to a DeepSet, especially is the simulation budget is high.

        The model consists of multiple stacked equivariant transformation modules followed by an invariant pooling
        module to produce a compact set representation.

        The equivariant layers perform many-to-many transformations, preserving structural information, while
        the final invariant module aggregates the set into a lower-dimensional summary.

        The model supports various activation functions, kernel initializations, and optional spectral normalization
        for stability. Pooling mechanisms can be specified for both intermediate and final aggregation steps.

        Parameters
        ----------
        summary_dim : int, optional
            Dimensionality of the final learned summary statistics. Default is 16.
        depth : int, optional
            Number of stacked equivariant modules. Default is 2.
        inner_pooling : str, optional
            Type of pooling operation applied within equivariant modules, such as "mean".
            Default is "mean".
        output_pooling : str, optional
            Type of pooling operation applied in the final invariant module, such as "mean".
            Default is "mean".
        mlp_widths_equivariant : Sequence[int], optional
            Widths of the MLP layers inside the equivariant modules. Default is (64, 64).
        mlp_widths_invariant_inner : Sequence[int], optional
            Widths of the inner MLP layers within the invariant module. Default is (64, 64).
        mlp_widths_invariant_outer : Sequence[int], optional
            Widths of the outer MLP layers within the invariant module. Default is (64, 64).
        mlp_widths_invariant_last : Sequence[int], optional
            Widths of the MLP layers in the final invariant transformation. Default is (64, 64).
        activation : str, optional
            Activation function used throughout the network, such as "gelu". Default is "gelu".
        kernel_initializer : str, optional
            Initialization strategy for kernel weights, such as "he_normal". Default is "he_normal".
        dropout : int, float, or None, optional
            Dropout rate applied within MLP layers. Default is 0.05.
        spectral_normalization : bool, optional
            Whether to apply spectral normalization to stabilize training. Default is False.
        **kwargs
            Additional keyword arguments passed to the equivariant and invariant modules.
        """

        super().__init__(**kwargs)

        # Stack of equivariant modules for a many-to-many learnable transformation
        self.equivariant_modules = []
        for _ in range(depth):
            equivariant_module = EquivariantLayer(
                mlp_widths_equivariant=mlp_widths_equivariant,
                mlp_widths_invariant_inner=mlp_widths_invariant_inner,
                mlp_widths_invariant_outer=mlp_widths_invariant_outer,
                activation=activation,
                kernel_initializer=kernel_initializer,
                spectral_normalization=spectral_normalization,
                dropout=dropout,
                pooling=inner_pooling,
            )
            self.equivariant_modules.append(equivariant_module)

        # Invariant module for a many-to-one transformation
        self.invariant_module = InvariantLayer(
            mlp_widths_inner=mlp_widths_invariant_last,
            mlp_widths_outer=mlp_widths_invariant_last,
            activation=activation,
            kernel_initializer=kernel_initializer,
            dropout=dropout,
            pooling=output_pooling,
            spectral_normalization=spectral_normalization,
        )

        # Output linear layer to project set representation down to "summary_dim" learned summary statistics
        self.output_projector = keras.layers.Dense(summary_dim, activation="linear", name="output_projector")

        self.summary_dim = summary_dim

    def call(self, x: Tensor, training: bool = False, **kwargs) -> Tensor:
        """
        Performs the forward pass of a hierarchical deep invariant transformation.

        This function applies a sequence of equivariant transformations to the input tensor,
        preserving structural relationships while refining representations. After passing
        through the equivariant modules, the data is processed by an invariant transformation,
        which aggregates information into a lower-dimensional representation. The final output
        is projected to the specified summary dimension using a linear layer.

        Parameters
        ----------
        x : Tensor
            Input tensor representing a set or collection of elements to be transformed.
        training : bool, optional
            Whether the model is in training mode, affecting layers like dropout. Default is False.
        **kwargs
            Additional keyword arguments passed to the transformation layers.

        Returns
        -------
        output : Tensor
            Transformed tensor with a reduced dimensionality, representing the learned summary
            of the input set.
        """

        for em in self.equivariant_modules:
            x = em(x, training=training)

        x = self.invariant_module(x, training=training)

        return self.output_projector(x)



import keras

from bayesflow.types import Shape, Tensor
from bayesflow.utils import layer_kwargs
from bayesflow.utils.serialization import serializable, deserialize


@serializable
class Distribution(keras.Layer):
    def __init__(self, **kwargs):
        super().__init__(**layer_kwargs(kwargs))

    def call(self, samples: Tensor) -> Tensor:
        return keras.ops.exp(self.log_prob(samples))

    def log_prob(self, samples: Tensor, *, normalize: bool = True) -> Tensor:
        raise NotImplementedError

    def sample(self, batch_shape: Shape) -> Tensor:
        raise NotImplementedError

    def compute_output_shape(self, input_shape: Shape) -> Shape:
        return keras.ops.shape(self.sample(input_shape[0:1]))

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))



import numpy as np
from scipy.special import expit

from .benchmark_simulator import BenchmarkSimulator


class BernoulliGLMRaw(BenchmarkSimulator):
    def __init__(self, T: int = 100, rng: np.random.Generator = None):
        """Bernoulli GLM raw simulated benchmark.
        See: https://arxiv.org/pdf/2101.04653.pdf, Task T.6

        Parameters
        ----------
        T: int, optional, default: 100
            The simulated duration of the task (eq. the number of Bernoulli draws).
        rng: np.random.Generator or None, optional, default: None
            An optional random number generator to use.
        """

        self.T = T
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

        # Covariance matrix computed once for efficiency
        F = np.zeros((9, 9))
        i = np.arange(9)
        F[i, i] = 1 + np.sqrt(i / 9)
        F[i[1:], i[:-1]] = -2
        F[i[2:], i[:-2]] = 1
        self.cov = np.linalg.inv(F.T @ F)

    def prior(self):
        """Generates a random draw from the custom prior over the 10
        Bernoulli GLM parameters (1 intercept and 9 weights). Uses a
        global covariance matrix `Cov` for the multivariate Gaussian prior
        over the model weights, which is pre-computed for efficiency.

        Returns
        -------
        params : np.ndarray of shape (10, )
            A single draw from the prior.
        """

        beta = self.rng.normal(0, 2)
        f = self.rng.multivariate_normal(np.zeros(9), self.cov)
        return np.append(beta, f)

    def observation_model(self, params: np.ndarray):
        """Simulates data from the custom Bernoulli GLM likelihood and
        returns the raw Bernoulli data.

        Parameters
        ----------
        params : np.ndarray of shape (10, )
            The vector of model parameters (`params[0]` is intercept, `params[i], i > 0` are weights)

        Returns
        -------
        x : np.ndarray of shape (T, 10)
            The full simulated set of Bernoulli draws and design matrix.
            Should be configured with an additional trailing dimension if the data is (properly) to be treated as a set.
        """

        # Unpack parameters
        beta, f = params[0], params[1:]

        # Generate design matrix
        V = self.rng.normal(size=(9, self.T))

        # Draw from Bernoulli GLM and return
        z = self.rng.binomial(n=1, p=expit(V.T @ f + beta))
        return np.c_[np.expand_dims(z, axis=-1), V.T]



from collections.abc import Callable, Sequence
from functools import wraps
import inspect
from typing import overload, TypeVar
from bayesflow.types import Shape

Fn = TypeVar("Fn", bound=Callable[..., any])

# this can be done better, but not compactly in Python < 3.12
Decorator = Fn


def allow_args(fn: Decorator) -> Decorator:
    """Decorator to allow another decorator to be called with or without arguments."""

    @overload
    def wrapper(f: Fn) -> Fn: ...
    @overload
    def wrapper(*fargs: any, **fkwargs: any) -> Fn: ...
    def wrapper(*fargs: any, **fkwargs: any) -> Fn:
        if len(fargs) == 1 and not fkwargs and callable(fargs[0]):
            # called without arguments
            return fn(fargs[0])
        else:
            # called with arguments, bind
            return lambda f: fn(f, *fargs, **fkwargs)

    return wrapper


def alias(*aliases: str) -> Decorator:
    """Decorator to create aliases for keyword arguments"""
    aliases = list(set(aliases))

    def alias_wrapper(fn: Fn) -> Fn:
        nonlocal aliases

        signature = inspect.signature(fn)
        parameter_names = list(signature.parameters.keys())
        candidates = [name for name in aliases if name in parameter_names]

        if not candidates:
            raise ValueError("Found no valid argument candidates in the alias list.")
        if len(candidates) > 1:
            raise ValueError(f"Found multiple valid argument candidates in the alias list: {candidates!r}")

        argname = candidates[0]
        argpos = parameter_names.index(argname)

        aliases.remove(argname)

        del signature
        del parameter_names
        del candidates

        @wraps(fn)
        def wrapper(*args, **kwargs):
            # check if multiple aliases are specified
            matches = [name for name in kwargs if name in aliases]

            if not matches:
                return fn(*args, **kwargs)

            if len(matches) > 1 or (len(matches) > 0 and len(args) > argpos):
                raise TypeError(
                    f"{fn.__name__}() got multiple values for argument {argname!r}.\n"
                    f"This argument is also aliased as {aliases!r}"
                )

            # map aliases to base name
            kwargs[argname] = kwargs.pop(matches[0])

            return fn(*args, **kwargs)

        return wrapper

    return alias_wrapper


def argument_callback(argname: str, callback: Callable[[any], any]) -> Decorator:
    """Decorator to apply a callback to an argument before passing it to the function"""

    def callback_wrapper(fn: Fn) -> Fn:
        argpos = list(inspect.signature(fn).parameters.keys()).index(argname)

        @wraps(fn)
        def wrapper(*args, **kwargs):
            if argname in kwargs:
                kwargs[argname] = callback(kwargs[argname])
            elif len(args) > argpos:
                args = list(args)
                args[argpos] = callback(args[argpos])

            return fn(*args, **kwargs)

        return wrapper

    return callback_wrapper


def allow_batch_size(fn: Callable):
    """Decorator to allow an integer batch_size argument in addition to a tuple batch_shape argument"""

    def callback(x):
        if isinstance(x, Sequence):
            return x

        return (x,)

    fn = argument_callback("batch_shape", callback)(fn)
    fn = alias("batch_shape", "batch_size")(fn)

    return fn


def sanitize_input_shape(fn: Callable):
    """Decorator to replace the first dimension in ..._shape arguments with a dummy batch size if it is None"""

    # The Keras functional API passes input_shape = (None, second_dim, third_dim, ...), which
    # causes problems when constructions like self.call(keras.ops.zeros(input_shape)) are used
    # in build. To alleviate those problems, this decorator replaces None with an arbitrary batch size.
    def callback(input_shape: Shape) -> Shape:
        if input_shape[0] is None:
            input_shape = list(input_shape)
            input_shape[0] = 1
            return tuple(input_shape)
        return input_shape

    args = inspect.getfullargspec(fn).args
    for arg in args:
        if arg.endswith("_shape"):
            fn = argument_callback(arg, callback)(fn)
    return fn



from collections.abc import Sequence, Mapping

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from scipy.stats import binom

from bayesflow.utils import logging
from bayesflow.utils import prepare_plot_data, add_titles_and_labels, prettify_subplots


def calibration_histogram(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    figsize: Sequence[float] = None,
    num_bins: int = 10,
    binomial_interval: float = 0.99,
    label_fontsize: int = 16,
    title_fontsize: int = 18,
    tick_fontsize: int = 12,
    color: str = "#132a70",
    num_col: int = None,
    num_row: int = None,
) -> plt.Figure:
    """Creates and plots publication-ready histograms of rank statistics for simulation-based calibration
    (SBC) checks according to [1].

    Any deviation from uniformity indicates miscalibration and thus poor convergence
    of the networks or poor combination between generative model / networks.

    [1] Talts, S., Betancourt, M., Simpson, D., Vehtari, A., & Gelman, A. (2018).
    Validating Bayesian inference algorithms with simulation-based calibration.
    arXiv preprint arXiv:1804.06788.

    Parameters
    ----------
    estimates      : np.ndarray of shape (n_data_sets, n_post_draws, n_params)
        The posterior draws obtained from n_data_sets
    targets     : np.ndarray of shape (n_data_sets, n_params)
        The prior draws obtained for generating n_data_sets
    variable_keys       : list or None, optional, default: None
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names    : list or None, optional, default: None
        The parameter names for nice plot titles. Inferred if None
    figsize          : tuple or None, optional, default : None
        The figure size passed to the matplotlib constructor. Inferred if None
    num_bins          : int, optional, default: 10
        The number of bins to use for each marginal histogram
    binomial_interval : float in (0, 1), optional, default: 0.99
        The width of the confidence interval for the binomial distribution
    label_fontsize    : int, optional, default: 16
        The font size of the y-label text
    title_fontsize    : int, optional, default: 18
        The font size of the title text
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels
    color        : str, optional, default '#a34f4f'
        The color to use for the histogram body
    num_row             : int, optional, default: None
        The number of rows for the subplots. Dynamically determined if None.
    num_col             : int, optional, default: None
        The number of columns for the subplots. Dynamically determined if None.

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving

    Raises
    ------
    ShapeError
        If there is a deviation form the expected shapes of `estimates` and `targets`.
    """

    plot_data = prepare_plot_data(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
        num_col=num_col,
        num_row=num_row,
        figsize=figsize,
    )

    estimates = plot_data.pop("estimates")
    targets = plot_data.pop("targets")

    # Determine the ratio of simulations to prior draw
    # num_params = plot_data['num_variables']
    num_sims = estimates.shape[0]
    num_draws = estimates.shape[1]

    ratio = int(num_sims / num_draws)

    # Log a warning if N/B ratio recommended by Talts et al. (2018) < 20
    if ratio < 20:
        logging.warning(
            "The ratio of simulations / posterior draws should be > 20 "
            f"for reliable variance reduction, but your ratio is {ratio}. "
            "Confidence intervals might be unreliable!"
        )

    # Set num_bins automatically, if nothing provided
    if num_bins is None:
        num_bins = int(ratio / 2)
        # Attempt a fix if a single bin is determined so plot still makes sense
        if num_bins == 1:
            num_bins = 4

    # Compute ranks (using broadcasting)
    ranks = np.sum(estimates < targets[:, np.newaxis, :], axis=1)

    # Compute confidence interval and mean
    num_trials = int(targets.shape[0])
    # uniform distribution expected -> for all bins: equal probability
    # p = 1 / num_bins that a rank lands in that bin
    endpoints = binom.interval(binomial_interval, num_trials, 1 / num_bins)
    mean = num_trials / num_bins  # corresponds to binom.mean(N, 1 / num_bins)

    for j, ax in enumerate(plot_data["axes"].flat):
        ax.axhspan(endpoints[0], endpoints[1], facecolor="gray", alpha=0.3)
        ax.axhline(mean, color="gray", zorder=0, alpha=0.9)
        sns.histplot(ranks[:, j], kde=False, ax=ax, color=color, bins=num_bins, alpha=0.95)
        ax.get_yaxis().set_ticks([])
    prettify_subplots(plot_data["axes"], tick_fontsize)

    add_titles_and_labels(
        axes=plot_data["axes"],
        num_row=plot_data["num_row"],
        num_col=plot_data["num_col"],
        title=plot_data["variable_names"],
        xlabel="Rank statistic",
        ylabel="",
        title_fontsize=title_fontsize,
        label_fontsize=label_fontsize,
    )
    plot_data["fig"].tight_layout()

    return plot_data["fig"]



import keras

from .log_sinkhorn import log_sinkhorn
from .sinkhorn import sinkhorn

methods = {
    "sinkhorn": sinkhorn,
    "sinkhorn_knopp": sinkhorn,
    "log_sinkhorn": log_sinkhorn,
    "log_sinkhorn_knopp": log_sinkhorn,
}


def optimal_transport(x1, x2, method="log_sinkhorn", return_assignments=False, **kwargs):
    """Matches elements from x2 onto x1, such that the transport cost between them is minimized, according to the method
    and cost matrix used.

    Depending on the method used, elements in either tensor may be permuted, dropped, duplicated, or otherwise modified,
    such that the assignment is optimal.

    Note: this is just a dispatch function that calls the appropriate optimal transport method.
    See the documentation of the respective method for more details.

    :param x1: Tensor of shape (n, ...)
        Samples from the first distribution.

    :param x2: Tensor of shape (m, ...)
        Samples from the second distribution.

    :param method: Method used to compute the transport cost.
        Default: 'log_sinkhorn'

    :param return_assignments: Whether to return the assignment indices.
        Default: False

    :param kwargs: Additional keyword arguments that are passed to the optimization method.

    :return: Tensors of shapes (n, ...) and (m, ...)
        x1 and x2 in optimal transport permutation order.
    """
    assignments = methods[method.lower()](x1, x2, **kwargs)
    x2 = keras.ops.take(x2, assignments, axis=0)

    if return_assignments:
        return x1, x2, assignments

    return x1, x2



import keras

from functools import singledispatch


@singledispatch
def find_recurrent_net(arg, *args, **kwargs):
    raise TypeError(f"Cannot infer network from {arg!r}.")


@find_recurrent_net.register
def _(name: str, *args, **kwargs):
    match name.lower():
        case "lstm":
            constructor = keras.layers.LSTM
        case "gru":
            constructor = keras.layers.GRU
        case other:
            raise ValueError(f"Unsupported network name: '{other}'.")

    return constructor


@find_recurrent_net.register
def _(network: keras.Layer, *args, **kwargs):
    return network



"""
A collection of utility functions, mostly used for internal purposes.
"""

from . import (
    keras_utils,
    logging,
    numpy_utils,
)

from .callbacks import detailed_loss_callback

from .devices import devices

from .dict_utils import (
    convert_args,
    convert_kwargs,
    filter_kwargs,
    layer_kwargs,
    model_kwargs,
    sequential_kwargs,
    split_tensors,
    split_arrays,
    squeeze_inner_estimates_dict,
)

from .dispatch import (
    find_network,
    find_permutation,
    find_pooling,
    find_recurrent_net,
    find_summary_network,
    find_inference_network,
    find_distribution,
)

from .ecdf import simultaneous_ecdf_bands, ranks

from .functional import batched_call

from .git import (
    issue_url,
    pull_url,
    repo_url,
)

from .hparam_utils import find_batch_size, find_memory_budget

from .integrate import (
    integrate,
)

from .io import (
    pickle_load,
    format_bytes,
    parse_bytes,
)

from .jacobian import (
    jacobian,
    jacobian_trace,
    jvp,
    vjp,
)

from .optimal_transport import optimal_transport

from .plot_utils import (
    check_estimates_prior_shapes,
    prepare_plot_data,
    add_titles_and_labels,
    prettify_subplots,
    make_quadratic,
    add_metric,
)
from .serialization import serialize_value_or_type, deserialize_value_or_type

from .tensor_utils import (
    concatenate_valid,
    expand,
    expand_as,
    expand_to,
    expand_left,
    expand_left_as,
    expand_left_to,
    expand_right,
    expand_right_as,
    expand_right_to,
    expand_tile,
    pad,
    searchsorted,
    size_of,
    stack_valid,
    tile_axis,
    tree_concatenate,
    tree_stack,
    fill_triangular_matrix,
    weighted_mean,
)

from .classification import calibration_curve, confusion_matrix

from .validators import check_lengths_same

from ._docs import _add_imports_to_all

_add_imports_to_all(include_modules=["keras_utils", "logging", "numpy_utils"])



from collections.abc import Sequence, Mapping, Callable

import numpy as np

from ...utils.dict_utils import dicts_to_arrays


def calibration_error(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    resolution: int = 20,
    aggregation: Callable = np.median,
    min_quantile: float = 0.005,
    max_quantile: float = 0.995,
) -> dict[str, any]:
    """
    Computes an aggregate score for the marginal calibration error over an ensemble of approximate
    posteriors. The calibration error is given as the aggregate (e.g., median) of the absolute deviation
    between an alpha-CI and the relative number of inliers from ``estimates`` over multiple alphas in
    (0, 1).

    Parameters
    ----------
    estimates  : np.ndarray of shape (num_datasets, num_draws, num_variables)
        The random draws from the approximate posteriors over ``num_datasets``
    targets : np.ndarray of shape (num_datasets, num_variables)
        The corresponding ground-truth values sampled from the prior
    variable_keys : Sequence[str], optional (default = None)
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names : Sequence[str], optional (default = None)
        Optional variable names to show in the output.
    resolution    : int, optional, default: 20
        The number of credibility intervals (CIs) to consider
    aggregation   : callable or None, optional, default: np.median
        The function used to aggregate the marginal calibration errors.
        If ``None`` provided, the per-alpha calibration errors will be returned.
    min_quantile  : float in (0, 1), optional, default: 0.005
        The minimum posterior quantile to consider.
    max_quantile  : float in (0, 1), optional, default: 0.995
        The maximum posterior quantile to consider.

    Returns
    -------
    result : dict
        Dictionary containing:

        - "values" : float or np.ndarray
            The aggregated calibration error per variable
        - "metric_name" : str
            The name of the metric ("Calibration Error").
        - "variable_names" : str
            The (inferred) variable names.
    """

    samples = dicts_to_arrays(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
    )

    # Define alpha values and the corresponding quantile bounds
    alphas = np.linspace(start=min_quantile, stop=max_quantile, num=resolution)
    regions = 1 - alphas
    lowers = regions / 2
    uppers = 1 - lowers

    # Compute quantiles for each alpha, for each dataset and parameter
    quantiles = np.quantile(samples["estimates"], [lowers, uppers], axis=1)

    # Shape: (2, resolution, num_datasets, num_params)
    lower_bounds, upper_bounds = quantiles[0], quantiles[1]

    # Compute masks for inliers
    lower_mask = lower_bounds <= samples["targets"][None, ...]
    upper_mask = upper_bounds >= samples["targets"][None, ...]

    # Logical AND to identify inliers for each alpha
    inlier_id = np.logical_and(lower_mask, upper_mask)

    # Compute the relative number of inliers for each alpha
    alpha_pred = np.mean(inlier_id, axis=1)

    # Calculate absolute error between predicted inliers and alpha
    absolute_errors = np.abs(alpha_pred - alphas[:, None])

    # Aggregate errors across alpha
    error = aggregation(absolute_errors, axis=0)

    variable_names = samples["estimates"].variable_names
    return {"values": error, "metric_name": "Calibration Error", "variable_names": variable_names}



from collections.abc import Callable

import keras

from bayesflow.types import Tensor
from .jacobian import jacobian
from .vjp import vjp


def jacobian_trace(
    f: Callable[[Tensor], Tensor],
    x: Tensor,
    max_steps: int = None,
    return_output: bool = False,
    seed: int | keras.random.SeedGenerator = None,
):
    """Compute or estimate the trace of the Jacobian matrix of f.

    Parameters
    ----------
    f : callable
        The function to be differentiated.
    x :  Tensor of shape (n, ..., d)
        The input tensor to f.
    max_steps : int, optional
        The maximum number of steps to use for the estimate.
        If this does not exceed the dimensionality of f(x), use Hutchinson's algorithm to
        return an unbiased estimate of the Jacobian trace.
        Otherwise, perform an exact computation.
        Default: None
    return_output : bool, optional
        Whether to return the output of f(x) along with the trace of the Jacobian.
        Default: False
    seed : int or keras SeedGenerator, optional
        The seed to use for hutchinson trace estimation. Only has an effect when max_steps < d.

    Returns
    -------
    2-tuple of tensors:
        1. The output of f(x) (if return_output is True)
        2. Tensor of shape (n,)
            An unbiased estimate or the exact trace of the Jacobian of f.
    """
    dims = keras.ops.shape(x)[-1]

    if max_steps is None or dims <= max_steps:
        fx, jac = jacobian(f, x, return_output=True)
        trace = keras.ops.trace(jac, axis1=-2, axis2=-1)
    else:
        fx, trace = _hutchinson(f, x, steps=max_steps, return_output=True, seed=seed)

    if return_output:
        return fx, trace

    return trace


def _hutchinson(
    f: callable, x: Tensor, steps: int = 1, return_output: bool = False, seed: int | keras.random.SeedGenerator = None
):
    """Estimate the trace of the Jacobian matrix of f using Hutchinson's algorithm.

    :param f: The function to be differentiated.

    :param x: Tensor of shape (n,..., d)
        The input tensor to f.

    :param steps: The number of steps to use for the estimate.
        Higher values yield better precision.
        Default: 1

    :return: 2-tuple of tensors:
        1. The output of f(x)
        2. Tensor of shape (n,)
            An unbiased estimate of the trace of the Jacobian matrix of f.
    """
    shape = keras.ops.shape(x)
    trace = keras.ops.zeros(shape[:-1])

    fx, vjp_fn = vjp(f, x, return_output=True)

    for _ in range(steps):
        projector = keras.random.normal(shape, seed=seed)
        trace += keras.ops.sum(vjp_fn(projector) * projector, axis=-1)

    if return_output:
        return fx, trace

    return trace



import numpy as np

from bayesflow.utils.serialization import serializable, deserialize


@serializable
class ElementwiseTransform:
    """Base class on which other transforms are based"""

    def __call__(self, data: np.ndarray, inverse: bool = False, **kwargs) -> np.ndarray:
        if inverse:
            return self.inverse(data, **kwargs)

        return self.forward(data, **kwargs)

    @classmethod
    def from_config(cls, config: dict, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self) -> dict:
        raise NotImplementedError

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        raise NotImplementedError

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        raise NotImplementedError



import pytest
import numpy as np


@pytest.fixture()
def median_score():
    from bayesflow.scores import MedianScore

    return MedianScore()


@pytest.fixture()
def median_score_subnet():
    from bayesflow.scores import MedianScore

    return MedianScore(subnets=dict(value="mlp"))


@pytest.fixture()
def mean_score():
    from bayesflow.scores import MeanScore

    return MeanScore()


@pytest.fixture()
def normed_diff_score():
    from bayesflow.scores import NormedDifferenceScore

    return NormedDifferenceScore(k=3)


@pytest.fixture(scope="function")
def quantile_score():
    from bayesflow.scores import QuantileScore

    return QuantileScore(q=[0.2, 0.3, 0.4, 0.5, 0.7])


@pytest.fixture()
def multivariate_normal_score():
    from bayesflow.scores import MultivariateNormalScore

    return MultivariateNormalScore()


@pytest.fixture(
    params=[
        "median_score",
        "median_score_subnet",
        "mean_score",
        "normed_diff_score",
        "quantile_score",
        "multivariate_normal_score",
    ],
)
def scoring_rule(request):
    return request.getfixturevalue(request.param)


@pytest.fixture(scope="function")
def point_inference_network(scoring_rule):
    from bayesflow.networks import PointInferenceNetwork

    return PointInferenceNetwork(
        scores=dict(dummy_name=scoring_rule),
    )


@pytest.fixture(scope="function")
def quantile_point_inference_network():
    from bayesflow.networks import PointInferenceNetwork
    from bayesflow.scores import QuantileScore

    return PointInferenceNetwork(
        scores=dict(quantiles=QuantileScore(q=np.array([0.1, 0.4, 0.5, 0.7]), subnets=dict(value="mlp"))),
    )



import math

import numpy as np

import keras
from keras import ops

from bayesflow.types import Shape, Tensor
from bayesflow.utils.decorators import allow_batch_size
from bayesflow.utils.serialization import serializable, serialize

from .distribution import Distribution


@serializable
class DiagonalNormal(Distribution):
    """Implements a backend-agnostic diagonal Gaussian distribution."""

    def __init__(
        self,
        mean: int | float | np.ndarray | Tensor = 0.0,
        std: int | float | np.ndarray | Tensor = 1.0,
        trainable_parameters: bool = False,
        seed_generator: keras.random.SeedGenerator = None,
        **kwargs,
    ):
        """
        Initializes a backend-agnostic diagonal Gaussian distribution with optional learnable parameters.

        This class represents a Gaussian distribution with a diagonal covariance matrix, allowing for efficient
        sampling and density evaluation.

        The mean and standard deviation can be specified as fixed values or learned during training. The class also
        supports random number generation with an optional seed for reproducibility.

        Parameters
        ----------
        mean : int, float, np.ndarray, or Tensor, optional
            The mean of the Gaussian distribution. Can be a scalar or a tensor. Default is 0.0.
        std : int, float, np.ndarray, or Tensor, optional
            The standard deviation of the Gaussian distribution. Can be a scalar or a tensor.
            Default is 1.0.
        trainable_parameters : bool, optional
            Whether to treat the mean and standard deviation as learnable parameters. Default is False.
        seed_generator : keras.random.SeedGenerator, optional
            A Keras seed generator for reproducible random sampling. If None, a new seed
            generator is created. Default is None.
        **kwargs
            Additional keyword arguments passed to the base `Distribution` class.

        """

        super().__init__(**kwargs)
        self.mean = mean
        self.std = std

        self.trainable_parameters = trainable_parameters
        self.seed_generator = seed_generator or keras.random.SeedGenerator()

        self.dim = None
        self.log_normalization_constant = None
        self._mean = None
        self._std = None

    def build(self, input_shape: Shape) -> None:
        if self.built:
            return

        self.dim = int(input_shape[-1])

        self.mean = ops.cast(ops.broadcast_to(self.mean, (self.dim,)), "float32")
        self.std = ops.cast(ops.broadcast_to(self.std, (self.dim,)), "float32")

        self.log_normalization_constant = -0.5 * self.dim * math.log(2.0 * math.pi) - ops.sum(ops.log(self.std))

        if self.trainable_parameters:
            self._mean = self.add_weight(
                shape=ops.shape(self.mean),
                initializer=keras.initializers.get(self.mean),
                dtype="float32",
                trainable=True,
            )
            self._std = self.add_weight(
                shape=ops.shape(self.std), initializer=keras.initializers.get(self.std), dtype="float32", trainable=True
            )
        else:
            self._mean = self.mean
            self._std = self.std

    def log_prob(self, samples: Tensor, *, normalize: bool = True) -> Tensor:
        result = -0.5 * ops.sum((samples - self._mean) ** 2 / self._std**2, axis=-1)

        if normalize:
            result += self.log_normalization_constant

        return result

    @allow_batch_size
    def sample(self, batch_shape: Shape) -> Tensor:
        return self._mean + self._std * keras.random.normal(shape=batch_shape + (self.dim,), seed=self.seed_generator)

    def get_config(self):
        base_config = super().get_config()

        config = {
            "mean": self.mean,
            "std": self.std,
            "trainable_parameters": self.trainable_parameters,
            "seed_generator": self.seed_generator,
        }

        return base_config | serialize(config)
